import time
from qtpy.QtWidgets import (QWidget, QCheckBox, QVBoxLayout,
                            QLineEdit, QLabel, QPushButton, QApplication, QHBoxLayout,
                            QListWidget, QSizePolicy, QTabWidget)
from micromorph.batch.batch_functions import run_omnipose_on_folder, run_analysis_on_folder, run_measure360_on_folder
try:
    from cellpose_omni.models import MODEL_NAMES
except:
    MODEL_NAMES = ["bact_fluor_omni", "bact_phase_omni"]
from ._gui_utils import DropdownMenu, LabelledIntField, LabelledFloatField
from ._MaskFilterWidget import MaskFilterWidget
from superqt import QCollapsible
from micromorph.defaults.defaults import analysis_options, filter_options, export_options, segmentation_options, image_loading_options


class _loadingSettings(QWidget):
    def __init__(self, parent: QWidget):
        super().__init__()
        self.setWindowTitle("Image Loading Settings")
        self._parent = parent

        self.image_reader = DropdownMenu("Image Reader", ["PIL", "tifffile"])
        self.split_axis = DropdownMenu("Split Axis", ["None", "0", "1"])
        self.split_channel = DropdownMenu("Split Channel", ["1", "2"])

        layout = QVBoxLayout()
        layout.addWidget(self.image_reader)
        layout.addWidget(self.split_axis)
        layout.addWidget(self.split_channel)
        self.setLayout(layout)

    def closeEvent(self, event):
        updated_options = {'image_reader': self.returnImageReader(),
                           'split_axis': self.returnSplitAxis(),
                           'split_channel': self.returnSplitChannel()}

        self._parent.image_loading_options.update(updated_options)


    def returnImageReader(self):
        return self.image_reader.returnSelected()


    def returnSplitAxis(self):
        selected_axis = self.split_axis.returnSelected()
        if selected_axis == "None":
            return None
        else:
            return int(selected_axis)


    def returnSplitChannel(self):
        return int(self.split_channel.returnSelected())


class _segmentationSettings(QWidget):
    def __init__(self, parent):
        super().__init__()
        self.setWindowTitle("Segmentation Settings")

        self._parent = parent

        self.params_widget = MaskFilterWidget(None)
        collapsible = QCollapsible("Mask filter options")
        collapsible.collapse(True)
        collapsible.addWidget(self.params_widget)
        collapsible.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        # dropdown menu
        self.model_name_dropdown = DropdownMenu("Model Name", MODEL_NAMES)
        idx_selected = MODEL_NAMES.index(self._parent.segmentation_options["model"])
        self.model_name_dropdown.cb.setCurrentIndex(idx_selected)
        self.model_name_dropdown.setContentsMargins(0, 0, 0, 0)

        self.gpu_checkbox = QCheckBox("Use GPU")
        self.gpu_checkbox.setChecked(self._parent.segmentation_options["use_gpu"])


        layout = QVBoxLayout()
        layout.addWidget(self.model_name_dropdown)
        layout.addWidget(self.gpu_checkbox)
        layout.addWidget(collapsible)

        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

    def closeEvent(self, event):
        updated_parameters = {'model': self.model_name_dropdown.returnSelected(),
                              'use_gpu': self.gpu_checkbox.isChecked(),
                              'binary_filter_options': self.params_widget.get_filter_options()}

        # update dictionary with settings
        self._parent.segmentation_options.update(updated_parameters)


class _filterSettings(QWidget):
    def __init__(self, parent, method):
        super().__init__()
        self._parent = parent
        self.method = method

        self.widgets = []

        self.layout = QVBoxLayout()
        self.layout.setSpacing(0)

        if method == "stdev":
            self.std_dev = LabelledIntField("Std Dev distance from Median", 1)
            self.layout.addWidget(self.std_dev)
            self.widgets.append(self.std_dev)
        elif method == "derivative":
            self.derivative = LabelledIntField("Derivative Threshold", 250)
            self.layout.addWidget(self.derivative)
            self.widgets.append(self.derivative)
        elif method == "sav-gol":
            self.window_size = LabelledIntField("Savitzky-Golay Window Size", 25)
            self.polyorder = LabelledIntField("Savitzky-Golay Polyorder", 3)

            self.layout.addWidget(self.window_size)
            self.layout.addWidget(self.polyorder)

            self.widgets.append(self.window_size)
            self.widgets.append(self.polyorder)

        self.layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(self.layout)

    def getValues(self):
        values = []
        for widget in self.widgets:
            values.append(widget.getValue())
        return values

    def removeWidget(self):
        print("Removing widgets")
        for widget in self.widgets:
            self.layout.removeWidget(widget)


class _analysisSettings(QWidget):
    def __init__(self, parent, batch_tools=True):
        super().__init__()
        self._parent = parent
        try:
            self._parent.analysis_options
        except AttributeError:
            # Populate with defaults
            self._parents.analysis_options = analysis_options()

        self.setWindowTitle("Analysis Settings")

        self.layout = QVBoxLayout()

        self.pxsize = LabelledFloatField("Pixel Size (nm/px)", self._parent.analysis_options['pxsize'])
        self.psfFWHM = LabelledFloatField("PSF FWHM (nm)", self._parent.analysis_options['psfFWHM'])
        self.n_lines = LabelledIntField("Number of lines", self._parent.analysis_options['n_lines'])

        self.filter_results = QCheckBox("Filter Results")
        self.filter_results.setChecked(self._parent.filter_options["apply_filter"])

        self.fit_type_options = ["fluorescence", "phase", "tophat"]
        self.fit_type_dropdown = DropdownMenu("Fit Type", self.fit_type_options)

        options = ["derivative", "stdev", "None"]

        self.filter_options = DropdownMenu("Filter Options", options)

        if batch_tools:
            # Export options, default both ticked
            self.export_csv = QCheckBox("Export CSV")
            self.export_pickle = QCheckBox("Export Pickle")
            self.export_hdf5 = QCheckBox("Export HDF5")
            self.export_csv.setChecked(True)
            self.export_pickle.setChecked(True)
            self.export_hdf5.setChecked(True)

        idx_selected = options.index(self._parent.filter_options['filter_type'])
        self.filter_options.cb.setCurrentIndex(idx_selected)
        self.filter_parameters = _filterSettings(self, self.filter_options.returnSelected())
        self.filter_options.cb.currentIndexChanged.connect(self.selectionchange)

        fit_type_selected_idx = self.fit_type_options.index(self._parent.analysis_options['fit_type'])
        self.fit_type_dropdown.cb.setCurrentIndex(fit_type_selected_idx)

        if batch_tools:
            export_label = QLabel("Export Options")
            export_label.setStyleSheet("font-weight: bold")
            self.layout.addWidget(export_label)
            self.layout.addWidget(self.export_csv)
            self.layout.addWidget(self.export_pickle)
            self.layout.addWidget(self.export_hdf5)

        self.tabs = QTabWidget()

        self.measure360_tab = QWidget()
        self.measure360_layout = QVBoxLayout()
        self.measure360_tab.setLayout(self.measure360_layout)
        self.measure360_layout.addWidget(self.n_lines)
        self.measure360_layout.addWidget(self.filter_results)
        self.measure360_layout.addWidget(self.filter_options)
        self.measure360_layout.addWidget(self.filter_parameters)
        self.measure360_layout.addStretch()


        self.full_analysis_tab = QWidget()
        self.full_analysis_layout = QVBoxLayout()
        self.full_analysis_tab.setLayout(self.full_analysis_layout)
        self.initialiseFullAnalysisTab()

        self.tabs.addTab(self.full_analysis_tab, "Full Analysis")
        self.tabs.addTab(self.measure360_tab, "Measure360")

        label_analysis = QLabel("Analysis Settings")
        label_analysis.setStyleSheet("font-weight: bold")

        self.layout.addWidget(label_analysis)
        self.layout.addWidget(self.pxsize)
        self.layout.addWidget(self.psfFWHM)
        self.layout.addWidget(self.fit_type_dropdown)
        self.layout.addWidget(self.tabs)

        self.setLayout(self.layout)

    def initialiseFullAnalysisTab(self):
        # Set parameters from the analysis_options dictionary
        self.n_widths = LabelledIntField("Number of Widths", self._parent.analysis_options['n_widths'])
        self.boundary_smoothing_factor = LabelledIntField("Boundary Smoothing Factor", self._parent.analysis_options['boundary_smoothing_factor'])
        self.error_threshold = LabelledFloatField("Error Threshold", self._parent.analysis_options['error_threshold'])
        self.max_iter = LabelledIntField("Max Iterations", self._parent.analysis_options['max_iter'])
        self.min_distance_to_boundary = LabelledIntField("Min Distance to Boundary", self._parent.analysis_options['min_distance_to_boundary'])
        self.step_size = LabelledFloatField("Step Size", self._parent.analysis_options['step_size'])
        self.spline_spacing = LabelledFloatField("Spline Spacing (px)", self._parent.analysis_options['spline_spacing'])
        self.spline_order = LabelledIntField("Spline Order", self._parent.analysis_options['spline_val'])


        self.show_boundaries_checkbox = QCheckBox("Show Boundaries")
        self.show_midlines_checkbox = QCheckBox("Show Midlines")

        self.full_analysis_layout.addWidget(self.n_widths)
        self.full_analysis_layout.addWidget(self.boundary_smoothing_factor)
        self.full_analysis_layout.addWidget(self.error_threshold)
        self.full_analysis_layout.addWidget(self.max_iter)
        self.full_analysis_layout.addWidget(self.min_distance_to_boundary)
        self.full_analysis_layout.addWidget(self.step_size)
        self.full_analysis_layout.addWidget(self.spline_spacing)
        self.full_analysis_layout.addWidget(self.spline_order)
        self.full_analysis_layout.addWidget(self.show_boundaries_checkbox)
        self.full_analysis_layout.addWidget(self.show_midlines_checkbox)

    def selectionchange(self, i):
        print(self.filter_options.returnSelected())
        self.layout.removeWidget(self.filter_parameters)
        self.filter_parameters = _filterSettings(self, self.filter_options.returnSelected())
        self.layout.addWidget(self.filter_parameters)

    # Print "close" when the window is closed
    def closeEvent(self, event):
        # Update parent.options

        updated_parameters = {'pxsize': self.pxsize.getValue(),
                              'psfFWHM': self.psfFWHM.getValue(),
                              'n_lines': self.n_lines.getValue(),
                              'n_widths': self.n_widths.getValue(),
                              'boundary_smoothing_factor': self.boundary_smoothing_factor.getValue(),
                              'error_threshold': self.error_threshold.getValue(),
                              'max_iter': self.max_iter.getValue(),
                              'min_distance_to_boundary': self.min_distance_to_boundary.getValue(),
                              'step_size': self.step_size.getValue(),
                              'fit_type': self.fit_type_dropdown.returnSelected(),
                              'show_boundaries': self.show_boundaries_checkbox.isChecked(),
                              'show_midlines': self.show_midlines_checkbox.isChecked(),
                              'spline_val': self.spline_order.getValue(),
                              'spline_spacing': self.spline_spacing.getValue()}

        self._parent.analysis_options.update(updated_parameters)

        updated_filter_options = {'apply_filter': self.filter_results.isChecked(),
                                  'filter_type': self.filter_options.returnSelected(),
                                  'filter_settings': self.filter_parameters.getValues()}

        self._parent.filter_options.update(updated_filter_options)

        if hasattr(self, 'export_csv'):
            self._parent.export_options.update({'export_type': self.check_output_type()})

    def check_output_type(self):
        # Check which checkboxes are ticked, if both are ticked, return 'both'
        if self.export_csv.isChecked() and self.export_pickle.isChecked() and self.export_hdf5.isChecked():
            return 'all'
        elif self.export_csv.isChecked():
            return 'csv'
        elif self.export_pickle.isChecked():
            return 'pickle'
        elif self.export_hdf5.isChecked():
            return 'hdf5'
        else:
            return 'none'


class _folderString(QWidget):
    def __init__(self, label, text):
        super().__init__()
        layout = QHBoxLayout()
        self.label = QLabel(label)
        self.edit_line = QLineEdit(text)
        layout.addWidget(self.label)
        layout.addWidget(self.edit_line)
        self.setLayout(layout)


class _checkboxPair(QWidget):
    def __init__(self, options, default):
        super().__init__()
        layout = QVBoxLayout()
        self.checkboxes = [QCheckBox(option) for option in options]
        for checkbox in self.checkboxes:
            layout.addWidget(checkbox)

        # Set first checkbox to checked
        self.checkboxes[default].setChecked(True)
        self.setLayout(layout)

        self.checkboxes[0].stateChanged.connect(lambda: self.set_status_to_opposite(0))
        self.checkboxes[1].stateChanged.connect(lambda: self.set_status_to_opposite(1))

    def set_status_to_opposite(self, a):
        # Get the status of the checkbox which just changed, set the other one to the opposite
        if self.checkboxes[a].isChecked():
            self.checkboxes[1-a].setChecked(False)
        else:
            self.checkboxes[1-a].setChecked(True)


class _batchtoolsPanel(QWidget):
    def __init__(self):
        super().__init__()
        layout = QVBoxLayout()

        self.input_directory = _folderString("Input Directory", "")
        self.output_directory = _folderString("Output Directory", "")

        # Create checkboxes
        segmentation_options_labels = ["Run Segmentation", "Load Segmentation"]
        self.segmentation_checkboxes = _checkboxPair(segmentation_options_labels, 0)

        analysis_options_labels = ["Full Analysis", "Measure360"]
        self.analysis_checkboxes = _checkboxPair(analysis_options_labels, 1)

        self.pil_image_reading_checkbox = QCheckBox("Use pil image reader.")

        buttons_layout = QHBoxLayout()
        self.segmentation_settings_button = QPushButton("Segmentation Settings")
        self.analysis_settings_button = QPushButton("Analysis Settings")
        buttons_layout.addWidget(self.segmentation_settings_button)
        buttons_layout.addWidget(self.analysis_settings_button)

        run_buttons_layout = QHBoxLayout()
        self.run_segmentation_button = QPushButton("Run Segmentation")
        self.run_analysis_button = QPushButton("Run Analysis")
        run_buttons_layout.addWidget(self.run_segmentation_button)
        run_buttons_layout.addWidget(self.run_analysis_button)

        self.run_button = QPushButton("Run")
        self.loading_settings_button = QPushButton("Image Loading Settings")

        self.add_button = QPushButton("Add Folder")
        self.add_button.clicked.connect(self.add_to_list)

        self.remove_button = QPushButton("Remove Folder")
        self.remove_button.clicked.connect(self.remove_from_list)

        self.listwidg = QListWidget()

        # ---- Add widgets to layout ----
        layout.addWidget(self.input_directory)
        layout.addWidget(self.add_button)
        layout.addWidget(self.remove_button)
        layout.addWidget(self.listwidg)
        layout.addWidget(self.segmentation_checkboxes)
        layout.addWidget(self.analysis_checkboxes)
        layout.addWidget(self.loading_settings_button)
        layout.addLayout(buttons_layout)
        layout.addLayout(run_buttons_layout)
        layout.addWidget(self.run_button)

        self.setLayout(layout)

        # ---- Set default options ----
        self.analysis_options = analysis_options()
        self.filter_options = filter_options()
        self.export_options = export_options()
        self.segmentation_options = segmentation_options()
        self.image_loading_options = image_loading_options()

        # ---- Connect operations ----
        self.analysis_settings_button.clicked.connect(self.open_analysis_settings)
        self.segmentation_settings_button.clicked.connect(self.open_segmentation_settings)
        self.run_analysis_button.clicked.connect(self.run_analysis)
        self.run_segmentation_button.clicked.connect(self.run_segmentation)
        self.run_button.clicked.connect(self.run_all)
        self.loading_settings_button.clicked.connect(self.open_loading_settings)

    def open_loading_settings(self):
        self.loading_settings = _loadingSettings(self)
        self.loading_settings.setStyleSheet(self.styleSheet())
        self.loading_settings.show()

    def open_analysis_settings(self):
        self.analysis_settings = _analysisSettings(self)
        self.analysis_settings.setStyleSheet(self.styleSheet())
        self.analysis_settings.show()

    def open_segmentation_settings(self):
        self.segmentation_settings = _segmentationSettings(self)
        self.segmentation_settings.setStyleSheet(self.styleSheet())
        self.segmentation_settings.show()

    def run_segmentation(self):
        folder_list = self.get_folder_list()

        for i, folder in enumerate(folder_list):
            # folder_name, model_name, gpu_option = False, chans = None, verbose = False
            print("Running segmentation on folder: " + folder)
            run_omnipose_on_folder(folder,
                                   self.segmentation_options["model"],
                                   self.segmentation_options["use_gpu"],
                                   chans=None,
                                   verbose=False,
                                   mask_filter_options=self.segmentation_options["binary_filter_options"],
                                   loading_method=self.image_loading_options["image_reader"],
                                   split_axis=self.image_loading_options["split_axis"],
                                   split_channel=self.image_loading_options["split_channel"])

    def run_analysis(self):
        folder_list = self.get_folder_list()

        for i, folder in enumerate(folder_list):
            # folder_name, options = dict()
            options = {'save': True,
                       'save_directory': folder,
                       'measure360': self.analysis_checkboxes.checkboxes[1].isChecked()}
            self.analysis_options.update(options)

            print("Running analysis on folder: " + folder)
            start_time = time.time()

            if self.analysis_options['measure360']:
                run_measure360_on_folder(folder,
                                         self.analysis_options,
                                         self.export_options,
                                         self.image_loading_options,
                                         self.filter_options)
            else:
                run_analysis_on_folder(folder,
                                       self.analysis_options)

            end_time = time.time()
            print("Analysis complete. Time elapsed: " + str(end_time - start_time) + " seconds")

    def run_all(self):
        if self.segmentation_checkboxes.checkboxes[0].isChecked():
            self.run_segmentation()
        self.run_analysis()

    def add_to_list(self):
        text = self.input_directory.edit_line.text()
        self.listwidg.addItem(text)

    def remove_from_list(self):
        self.listwidg.takeItem(self.listwidg.currentRow())

    def get_folder_list(self):
        # get all items in listwidg
        items = []
        for i in range(self.listwidg.count()):
            items.append(self.listwidg.item(i).text())

        return items
    

def open_micromorph_batch_gui():
    # set up app
    app = QApplication([])

    # Set app title
    app.setApplicationName("MicroMorph Batch")

    batch_tools_widget = _batchtoolsPanel()
    batch_tools_widget.show()

    app.exec_()


import numpy as np
import os
os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
# line above is a workaround for this problem : https://stackoverflow.com/questions/20554074/sklearn-omp-error-15-initializing-libiomp5md-dll-but-found-mk2iomp5md-dll-a
from skimage.measure import regionprops
import time
from typing import Union
from cellpose_omni import models
import logging

def run_omnipose(images: np.array, model_name: str, gpu_option: bool, chans: tuple or None =None, filter_options:
dict ={}) -> Union[np.array, np.array]:
    """
    Run the specified omnipose model on the input image(s)

    Parameters
    ----------
    images: np.ndarray
        The input image(s) to segment. If a single image is provided, it should be a 2D array. If multiple images are provided, they should be a 3D array with shape (n_images, height, width).
    model_name: str
        The name of the model to use.
    gpu_option: bool
        Whether to use the GPU for segmentation.
    chans: tuple or None
        The channels to use for segmentation. If None, the first channel will be used.
    filter_options: dict
        A dictionary containing the filter options for the segmentation. Related to the omnipose package. See this 
        link for more info: https://omnipose.readthedocs.io/examples/mono_channel_bact.html#run-segmentation
              
    Returns
    -------
    masks: np.ndarray
        The segmented masks. If a single image is provided, it will be a 2D array. If multiple images are provided, it will be a 3D array with shape (n_images, height, width).
    flows: np.ndarray
        The flow outputs. If a single image is provided, it will be a 2D array. If multiple images are provided, it will be a 3D array with shape (n_images, height, width).
    """
    # Should also add to this function the possibility to:
    model = models.CellposeModel(gpu=gpu_option, model_type=model_name)

    if chans is None:
        chans = [0, 0]  # If user didn't specify a channel, assume you need to use the first channel

    # TODO: params should be a dictionary which you can edit

    # define parameters - this is to become an input parameter
    params = {'channels': chans,  # always define this with the model
              'rescale': None,  # upscale or downscale your images, None = no rescaling
              'mask_threshold': 1,  # erode or dilate masks with higher or lower values -1
              'flow_threshold': 0.2,
              # default is .4, but only needed if there are spurious masks to clean up; slows down output
              'transparency': True,  # transparency in flow output
              'omni': True,  # we can turn off Omnipose mask reconstruction, not advised
              'cluster': True,  # use DBSCAN clustering
              'resample': True,  # whether to run dynamics on rescaled grid or original grid
              # 'verbose': False, # turn on if you want to see more output
              'tile': False,  # average the outputs from flipped (augmented) images; slower, usually not needed
              'niter': None,
              # None lets Omnipose calculate # of Euler iterations (usually <20) but you can tune it for over/under segmentation
              'augment': False,  # Can optionally rotate the image and average outputs, usually not needed
              'affinity_seg': False,  # new feature, stay tuned...
              }

    tic = time.time()
    if len(images.shape) == 2:
        try:
            masks, flows, styles = model.eval(images, **params)
        except ValueError as e:
            logging.error(f"Error in Omnipose segmentation: {e}")
            logging.info("Error in Omnipose segmentation. A blank mask will be returned.")
            masks = np.zeros_like(images)
            flows = np.zeros_like(images)
    else:
        n_images = len(images)
        n = range(n_images)

        masks_list, flows_list, styles = model.eval([images[i] for i in n], **params)

        # Convert list of masks to 3D array
        masks = np.dstack(masks_list)  # Transpose to match the original shape of the stack
        masks = np.moveaxis(masks, -1, 0)

        flows = np.copy(masks)

        for i, sublist in enumerate(flows_list):
            flows[i] = sublist[-1]  # -1

    net_time = time.time() - tic
    logging.info('Total segmentation time: {}s'.format(net_time))

    if filter_options:
        masks = filter_mask(masks, filter_options)

    return masks, flows


def filter_mask(mask:np.array, options: dict ={}) -> np.array:
    """
    Function to filter the segmentation mask basaed on basic morphological criteria such as area, length and width.

    Parameters
    ----------
    mask: np.ndarray
        The mask to filter. If a single mask is provided, it should be a 2D array. If multiple masks are provided, they should be a 3D array with shape (n_masks, height, width).
    options: dict
        A dictionary containing the filter options for the segmentation. The options are:
        - min_area: The minimum area of the bacteria.
        - max_area: The maximum area of the bacteria.
        - min_length: The minimum length of the bacteria.
        - max_length: The maximum length of the bacteria.
        - min_width: The minimum width of the bacteria.
        - max_width: The maximum width of the bacteria.

    Returns
    -------
    filtered_mask: np.ndarray
        The filtered mask. If a single mask is provided, it will be a 2D array. If multiple masks are provided, it will be a 3D array with shape (n_masks, height, width).
    """
    # Based on this: https://forum.image.sc/t/need-help-with-regionprops/30107/2

    # Get values from dictionary if they exist, otherwise use default values
    min_area = options.get("min_area", None)
    max_area = options.get("max_area", None)

    min_length = options.get("min_length", None)
    max_length = options.get("max_length", None)

    min_width = options.get("min_width", None)
    max_width = options.get("max_width", None)

    # Not sure if needed ---
    # label_mask = label(mask)
    label_mask = np.copy(mask)

    if len(label_mask.shape) == 2:
        # Start by getting regionprops
        regions = regionprops(label_mask)

        # Create a list to store the indices of the bacteria that meet the criteria
        bacteria_indices = []
        # Create a new mask with only the bacteria that meet the criteria
        filtered_mask = np.zeros_like(mask)

        # Loop through the regions
        for i, region in enumerate(regions):
            # Get the area, length and width of the region
            area = region.area
            length = region.axis_major_length
            width = region.axis_minor_length

            # Check if the region meets the criteria
            if (min_area is None or area >= min_area) and (max_area is None or area <= max_area) and \
                    (min_length is None or length >= min_length) and (max_length is None or length <= max_length) and \
                    (min_width is None or width >= min_width) and (max_width is None or width <= max_width):
                # If the region meets the criteria, add the index to the list
                bacteria_indices.append(region.label)

        for i in bacteria_indices:
            filtered_mask[label_mask == i] = i
    elif len(label_mask.shape) == 3:
        # Create stack to store the filtered masks
        filtered_mask = np.zeros_like(mask)

        # Loop through the frames
        for i, current_mask in enumerate(label_mask):
            current_filtered_mask = filter_mask(current_mask, options=options)

            filtered_mask[i] = current_filtered_mask
    else:
        raise ValueError("Mask must be 2D or 3D.")

    return filtered_mask



import numpy as np
import time
import logging
from scipy.signal import savgol_filter

from skimage.measure import profile_line, regionprops
from micromorph.bacteria.utilities import apply_mask_to_image

from micromorph.bacteria.phase_contrast_fitting import fit_phase_contrast_profile, fit_top_hat_profile
from micromorph.bacteria.fluorescence_fitting import fit_ring_profile

from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed

def process_single_bacterium(args):
    img, mask, i, bact, options = args
    img_processed = apply_mask_to_image(img, mask == bact.label)
    xc = bact.centroid[1]
    yc = bact.centroid[0]

    updated_options = {'magnitude': bact.major_axis_length * 1.5,
                       'area': bact.area,
                       'label': bact.label,
                       'bbox': bact.bbox}

    options.update(updated_options)

    current_bact = Bacteria360(img_processed, (yc, xc), options=options)

    return current_bact

def run_measure360(img: np.array, mask: np.array, options: dict = dict(), pool=None, close_pool=True):
    """
    Runs the measure360 function on a single image or a stack of images.

    Parameters
    ----------
    img : np.array
        Single frame (or stack), either phase contrast or fluorescence.
    mask : np.array
        Labelled mask from the image.
    options : dict, optional
        Dictionary of options to pass to the measure360 function.

    Returns
    -------
    all_bacteria : list
        List of Bacteria360 objects.
    """
    if pool is None:
        logging.info("Making a pool for multithreading - this only happens once!")
        pool = ThreadPoolExecutor()  # Create executor outside a with block
        logging.info("Pool created!")
    else:
        logging.info("Using existing pool for multithreading")

    all_bacteria = []
    if len(img.shape) == 2:
        # Get bacteria properties
        props = regionprops(mask)

        args = [(img, mask, i, bact, options) for i, bact in enumerate(props)]

        futures = [pool.submit(process_single_bacterium, arg) for arg in args]
        with tqdm(total=len(args)) as pbar:
            for future in as_completed(futures):
                result = future.result()
                all_bacteria.append(result)
                pbar.update(1)
    else:
        n_frames = img.shape[0]
        for i in range(n_frames):
            # Get ith image and mask
            img_current = np.copy(img[i])
            mask_current = np.copy(mask[i])

            current_data = run_measure360(img_current, mask_current, options=options, pool=pool, close_pool=False)

            for bact in current_data:
                bact.slice = i

            all_bacteria.extend(current_data)

    if close_pool:
        if pool:
            pool.shutdown(wait=True)

    return all_bacteria


class Bacteria360:
    """
    A class to store all the properties of a single bacterium.
    It requires an image and a mask to be initialised, which are used to calculate the properties.

    Parameters
    ----------
    img : np.array
        The image of the bacterium.
    centroid : tuple
        The centroid of the bacterium.
    options : dict
        Dictionary of options to pass to the measure360 function.

    Attributes
    ----------
    centroid : tuple
        The centroid of the bacterium.
    width_data : np.array
        The width data of the bacterium.
    width : float
        The minimum width of the bacterium.
    length : float
        The maximum width of the bacterium.
    area : float
        The area of the bacterium.
    label : int
        The label of the bacterium.
    bbox : tuple
        The bounding box of the bacterium.
    slice : int
        The slice of the stack the bacterium is from.
    n_lines : int
        The number of lines to measure the width along.
    magnitude : float
        The length of the lines.
    px_size : float
        The pixel size of the image.
    psfFWHM : float
        The FWHM of the PSF.
    xc : float
        The x-coordinate of the centroid.
    yc : float
        The y-coordinate of the centroid.
    xc_nm : float
        The x-coordinate of the centroid in nm.
    yc_nm : float
        The y-coordinate of the centroid in nm.

    """
    def __init__(self, img, centroid, options=dict()):

        n_lines = options.get('n_lines', 50)
        magnitude = options.get('magnitude', 100)
        pool = options.get('pool', None)
        area = options.get('area', 0)
        label = options.get('label', None)
        bbox = options.get('bbox', None)

        width_data = np.array(measure360(img, centroid, options=options))

        # remove any rows containing nan values
        width_data = width_data[~np.isnan(width_data).any(axis=1)]

        self.centroid = centroid
        self.width_data = width_data
        self.width = np.min(self.width_data[:, 1])
        self.length = np.max(self.width_data[:, 1])
        self.area = area
        self.label = label
        self.bbox = bbox
        self.slice = 0  # this is by default 0, but gets overwritten when data is generated from a stack

        self.n_lines = n_lines
        self.magnitude = magnitude
        self.px_size = options.get('px_size', None)
        self.psfFWHM = options.get('psfFWHM', None)

        self.xc = centroid[0]
        self.yc = centroid[1]

        self.xc_nm = self.xc * options.get('px_size', 1)
        self.yc_nm = self.yc * options.get('px_size', 1)


def fit_multiprocessing(img, angle, x_1, y_1, x_2, y_2, params):
    """
    Function to fit the width of the bacterium at a single angle.

    Parameters
    ----------
    img : np.array
        The image of the bacterium.
    angle : float
        The angle corresponding to the line. It is stores in the results array but has no bearing on the width
        calculation.
    x_1 : float
        The x-coordinate of the start of the line.
    y_1 : float
        The y-coordinate of the start of the line.
    x_2 : float
        The x-coordinate of the end of the line.
    y_2 : float
        The y-coordinate of the end of the line.
    params : list
        List of parameters [psfFWHM, pxsize, fit_type].

    Returns
    -------
    data : np.array
        Array of data [angle, width, x_1, y_1, x_2, y_2].
    """

    psfFWHM = params[0]
    pxsize = params[1]
    fit_type = params[2]

    # Set up array to collect data
    current_profile = profile_line(img, (y_1, x_1), (y_2, x_2))

    x = np.arange(0, len(current_profile)) * pxsize

    try:
        if fit_type == 'fluorescence':
            result = fit_ring_profile(x, current_profile, psfFWHM)
            width = result.params['R'] * 2
        elif fit_type == 'phase':
            result = fit_phase_contrast_profile(x, current_profile)
            width = result.params['width'].value
        elif fit_type == 'tophat':
            result = fit_top_hat_profile(x, current_profile)
            width = result.params['width'].value
        else:
            raise ValueError('Invalid fit type - please choose "fluorescence" or "phase".')
    except:
        width = np.nan

    if width < 0:
        width = np.nan

    data = np.array([angle, width, x_1, y_1, x_2, y_2])

    return data


def measure360(img, centroid, options=dict()):
    """
    This function measures the width of a bacterium at 360 degrees around its centroid.

    Parameters
    ----------
    img : np.array
        The image of the bacterium.
    centroid : tuple
        The centroid of the bacterium.
    options : dict
        Dictionary of options to pass to the measure360 function.

    Returns
    -------
    data : list
        List of data [angle, width, x_1, y_1, x_2, y_2].
    """
    n_lines = options.get('n_lines', 100)
    magnitude = options.get('magnitude', 100)
    psfFWHM = options.get('psfFWHM', 0.25)
    px_size = options.get('pxsize', 0.65)
    fit_type = options.get('fit_type', 'fluorescence')
    pool = options.get('pool', None)
    params = [psfFWHM, px_size, fit_type]

    # # Initialise multi processing pool
    # if pool is None:
    #     # Number of processes to run in parallel
    #     num_processes = multiprocessing.cpu_count()
    #     # Create a pool of worker processes
    #     pool = multiprocessing.Pool(processes=num_processes)

    # Define angle_range based on required number of lines
    angle_range = np.linspace(0, np.pi, n_lines)

    # Get points on circle
    x_1, y_1, x_2, y_2 = get_points_on_circle(centroid, magnitude/2, angle_range)

    # Check if any x or y is below 0 or above the image size
    x_1 = np.clip(x_1, 0, img.shape[1])
    y_1 = np.clip(y_1, 0, img.shape[0])
    x_2 = np.clip(x_2, 0, img.shape[1])
    y_2 = np.clip(y_2, 0, img.shape[0])

    data = []

    start = time.time()

    # possible way to run multiprocessing?
#     with multiprocessing.Pool() as pool:
#        data = pool.map(fit_multiprocessing, [(img, angle_range[i], x_1[i], y_1[i], x_2[i], y_2[i], params) for i in range(len(x_1))])
    # data = pool.starmap(fit_multiprocessing, [(img, angle_range[i], x_1[i], y_1[i], x_2[i], y_2[i], params) for i in range(len(x_1))])
        
    for i in range(len(x_1)):
        # logging.debug(f'Processing line {i+1} of {len(x_1)}')
        data.append(fit_multiprocessing(img, angle_range[i], x_1[i], y_1[i], x_2[i], y_2[i], params))

    if options.get('verbose', False):
        logging.debug('Time taken: ', time.time() - start)

    return data


def get_points_on_circle(centroid, magnitude, angle_range):
    """
    Function to get the points on a circle of a given radius around a centroid.

    Parameters
    ----------
    centroid : tuple
        The centroid of the bacterium.
    magnitude : float
        The radius of the circle.
    angle_range : np.array
        The range of angles to calculate the points.

    Returns
    -------
    x_1 : np.array
        The x-coordinates of the start of the lines.
    y_1 : np.array
        The y-coordinates of the start of the lines.
    x_2 : np.array
        The x-coordinates of the end of the lines.
    y_2 : np.array
        The y-coordinates of the end of the lines.
    """
    x_1 = centroid[1] + magnitude*np.cos(angle_range)
    y_1 = centroid[0] + magnitude*np.sin(angle_range)

    x_2 = centroid[1] - magnitude*np.cos(angle_range)
    y_2 = centroid[0] - magnitude*np.sin(angle_range)

    return x_1, y_1, x_2, y_2


def width_distribution(all_data):
    """
    Convenience function to plot the distribution of the widths of all the bacteria.

    Parameters
    ----------
    all_data : list
        List of Bacteria360 objects.

    Returns
    -------
    None
    """

    all_widths = []

    for data in all_data:
        width = data.width
        all_widths.append(width)

    plt.hist(all_widths)
    plt.show()
    print(len(all_widths))


def filter_measure360(bacteria, filter_type, filter_settings):
    """
    Function to filter the width data of a bacterium.

    Parameters
    ----------
    bacteria : Bacteria360
        The Bacteria360 object to be filtered.
    filter_type : str
        The type of filter to apply. Options are 'stdev', 'derivative' or 'sav-gol'.
    filter_settings : list
        The settings for the filter. For 'stdev' and 'derivative' this is a single value, for 'sav-gol' it is a list
        of two values [window, order].

    Returns
    -------
    bacteria : Bacteria360
        The Bacteria360 object with the filtered width data.
    """
    width_data = np.copy(bacteria.width_data[:, 1])

    if filter_type == 'stdev':
        # Filter the width
        idx = abs(width_data - np.median(width_data)) < filter_settings[0] * np.std(width_data)

        bacteria.width_data = bacteria.width_data[idx, :]
    elif filter_type == 'derivative':
        delta = np.diff(width_data)
        delta = np.insert(delta, 0, 0)
        idx = abs(delta) < filter_settings[0]

        bacteria.width_data = bacteria.width_data[idx, :]
    elif filter_type == 'sav-gol':
        window = filter_settings[0]
        order = filter_settings[1]
        width_data = savgol_filter(width_data, window, order)
        bacteria.width_data[:, 1] = width_data

    try:
        bacteria.width = np.min(bacteria.width_data[:, 1])
        bacteria.length = np.max(bacteria.width_data[:, 1])
    except:
        print("filtering didn't work")

    return bacteria




from skimage.io import imread
import matplotlib.pyplot as plt
import matplotlib as mpl
from omnipose.utils import rescale
import fastremap
import ncolor
import numpy as np

from micromorph import get_bacteria_list
from micromorph.segmentation import run_omnipose
from micromorph.measure360 import run_measure360

if __name__ == "__main__":
    # Load image and the corresponding mask
    image = imread(r"C:\Users\u1870329\Documents\GitHub\micromorph\micromorph\test-data\microcolony_omnisegger_example.tif")

    # We now generate the mask using Omnipose
    mask, _ = run_omnipose(image, 'bact_phase_omni', gpu_option=False, filter_options={'min_area': 50})
    # Now use the `get_bacteria_list` function to analyse all the bacteria in the image
    # bacteria_list = get_bacteria_list(image, mask, options={'pxsize': 110, 'fit_type': 'phase'})
    bacteria_list = run_measure360(image, mask, options={'n_angles': 50, 'pxsize': 65, 'fit_type': 'phase', 'psfFWHM': 250})



from skimage.io import imread
from micromorph import get_bacteria_list
from micromorph.measure360 import run_measure360

import matplotlib.pyplot as plt

if __name__ == "__main__":
    # Load image and the corresponding mask
    image_stack = imread(r"C:\Users\u1870329\Documents\GitHub\micromorph\micromorph\test-data\multiple-cells\example_stack.tif")
    mask_stack = imread(r"C:\Users\u1870329\Documents\GitHub\micromorph\micromorph\test-data\multiple-cells\example_stack_mask.tif")
    
    # Get measure360 values for the cells in the stack
    # bacteria_list = run_measure360(image_stack, mask_stack, options={'n_angles': 50, 'pxsize': 65, 'fit_type': 'phase', 'psfFWHM': 250})

    # or use get_bacteria_list to use "normal mode"
    bacteria_list = get_bacteria_list(image_stack, mask_stack, options={'pxsize': 65, 'n_widths': 5, 'fit_type': 'fluorescence', 'psfFWHM': 250})

    # Get widths and lengths
    widths = [bacteria.width for bacteria in bacteria_list]
    lengths = [bacteria.length for bacteria in bacteria_list]

    # Plot histograms
    fig, ax = plt.subplots(1, 2, figsize=(12, 5))
    ax[0].hist(widths, bins=30, color='blue', alpha=0.7)
    ax[0].set_title('Width Distribution')

    ax[1].hist(lengths, bins=30, color='green', alpha=0.7)
    ax[1].set_title('Length Distribution')
    plt.show()


from asyncio import as_completed
import time
import numpy as np
from skimage.measure import regionprops, label
from skimage.segmentation import clear_border
from .shape_analysis import (get_bacteria_length, get_bacteria_widths, get_bacteria_boundary, get_medial_axis,
                            smooth_medial_axis, extend_medial_axis)
from .utilities import apply_mask_to_image
import logging

from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed


class Bacteria:
    """
    A class to store all the properties of a single bacterium.

    Parameters
    ----------
    img: the image containg the bacterium
    mask: the mask of the bacterium
    options: a dictionary with options to be passed to the functions used to calculate the morphological properties of the bacterium

    Attributes
    ----------
    centroid: the centroid of the bacterium
    xc: the x-coordinate of the centroid
    yc: the y-coordinate of the centroid
    xc_nm: the x-coordinate of the centroid in nm
    yc_nm: the y-coordinate of the centroid in nm
    bbox: the bounding box of the bacterium
    axis_major_length: the major axis length of the bacterium
    axis_minor_length: the minor axis length of the bacterium
    orientation: the orientation of the bacterium
    area: the area of the bacterium
    boundary: the boundary of the bacterium
    medial_axis: the medial axis of the bacterium
    medial_axis_extended: the medial axis of the bacterium extended to the boundary
    slice: the slice of the stack where the bacterium is located (default is zero)
    all_widths: the widths of the bacterium at different points along the medial axis
    """

    def __init__(self, img, mask, options=dict()):
        # print("New version!!")
        pxsize = options.get('pxsize', 1)
        n_widths = options.get('n_widths', 5)
        boundary_smoothing_factor = options.get('boundary_smoothing_factor', 8)
        fit_type = options.get('fit_type', None)
        psfFWHM = options.get('psfFWHM', 0.250)
        error_threshold = options.get('error_threshold', 0.05)
        max_iter = options.get('max_iter', 50)
        min_distance_to_boundary = options.get('min_distance_to_boundary', 1)
        step_size = options.get('step_size', 1)

        # TODO: add to GUI!
        spline_spacing = options.get('spline_spacing', 0.25)
        spline_val = options.get('spline_val', 3)


        # Get properties from regionprops, then add them to the class.
        bacteria_props = regionprops(label(mask))

        self.centroid = np.asarray(bacteria_props[0].centroid)
        self.xc = bacteria_props[0].centroid[1]
        self.yc = bacteria_props[0].centroid[0]
        self.xc_nm = self.xc * pxsize
        self.yc_nm = self.yc * pxsize
        self.bbox = bacteria_props[0].bbox
        self.axis_major_length = bacteria_props[0].major_axis_length * pxsize
        self.axis_minor_length = bacteria_props[0].minor_axis_length * pxsize
        self.orientation = bacteria_props[0].orientation
        self.area = bacteria_props[0].area * pxsize**2

        # Apply filter to image to avoid possible issues when calculating width of bacteria in clumps.
        # also, crop image to reduce computational time of erosion etc
        bbox = self.bbox
        mask_cropped = np.copy(mask[bbox[0]-1:bbox[2]+1, bbox[1]-1:bbox[3]+1])
        image_cropped = np.copy(img[bbox[0]-1:bbox[2]+1, bbox[1]-1:bbox[3]+1])

        if fit_type == 'phase':
            # Apply mask to image
            img_filtered = apply_mask_to_image(image_cropped, mask_cropped, method='max')
        else:
            # Apply mask to image
            img_filtered = apply_mask_to_image(image_cropped, mask_cropped, method='min')

        # img_filtered = apply_mask_to_image(image_cropped, mask_cropped)

        # Calculate other properties with custom functions.
        try:
            self.boundary = get_bacteria_boundary(mask_cropped, boundary_smoothing_factor=boundary_smoothing_factor)
            self.perimeter = np.sum(np.sqrt(np.sum(np.diff(self.boundary, axis=0)**2, axis=1))) * pxsize
            self.circularity = (4 * np.pi * self.area) / (self.perimeter**2)

            start_time_medial_ax = time.perf_counter()
            self.medial_axis = smooth_medial_axis(get_medial_axis(mask_cropped), self.boundary,
                                                  error_threshold=error_threshold, max_iter=max_iter,
                                                  spline_val=spline_val, spline_spacing=spline_spacing)
            end_time_medial_ax = time.perf_counter()
            logging.debug(f"Time taken to calculate medial axis: {end_time_medial_ax - start_time_medial_ax:.2f} "
                          f"seconds")

            start_time_extend = time.perf_counter()
            self.medial_axis_extended = extend_medial_axis(self.medial_axis, self.boundary,
                                                           error_threshold=error_threshold, max_iter=max_iter,
                                                           min_distance_to_boundary=min_distance_to_boundary,
                                                           step_size=step_size)
            end_time_extend = time.perf_counter()
            logging.debug(f"Time taken to extend medial axis: {end_time_extend - start_time_extend:.2f} seconds")

            start_time_widths = time.perf_counter()
            logging.debug(f"Calculating widths using method {fit_type}")
            self.all_widths = get_bacteria_widths(img_filtered,
                                                  self.medial_axis,
                                                  n_lines=n_widths, pxsize=pxsize, fit_type=fit_type,
                                                  line_magnitude=bacteria_props[0].axis_minor_length*1.5,
                                                  psfFWHM=psfFWHM)
            end_time_widths = time.perf_counter()
            logging.debug(f"Time taken to calculate widths: {end_time_widths - start_time_widths:.2f} seconds")

            self.width = np.median(self.all_widths)
            if self.width < 0:
                self.width = None
            self.length = get_bacteria_length(self.medial_axis_extended, pxsize)

            # logging.debug(f"Width: {self.width:.2f} nm from {self.all_widths}")
            # logging.debug(f"Length: {self.length:.2f} nm")
            # TODO: improve how errors are dealt with
        except ValueError:
            logging.debug("Error calculating widths or length-VALUE")
            self.medial_axis = None
            self.medial_axis_extended = None
            self.all_widths = None
            self.width = None
            self.length = None
        except IndexError:
            logging.debug("Error calculating widths or length-INDEX")
            self.medial_axis = None
            self.medial_axis_extended = None
            self.all_widths = None
            self.width = None
            self.length = None

        # Set slice to zero, this gets updated if the image is part of a stack outside of this function.
        self.slice = 0


def get_bacteria_for_index(img: np.array, mask: np.array, i: int, options: dict):
    """
    Function which returns a Bacteria object for the i-th bacterium in the labelled mask.

    Parameters
    ----------
    img: np.array
        the image containing all the bacteria
    mask: np.array
        the labelled mask of the image
    i: int
        the index of the bacterium to be analysed
    options: dict
        a dictionary with options to be passed to the functions used to calculate the morphological properties of
        the bacterium

    Returns
    -------
    bac: Bacteria
        a Bacteria object containing the properties of the i-th bacterium
    """
    mask_current = np.copy(mask) == i

    bac = Bacteria(img, mask_current, options)
    bac.label = i

    return bac



def _process_bacterium(args):
    """Convenience function used for multi-processing."""
    try:
        img, mask, i, options = args
        return get_bacteria_for_index(img, mask, i, options)
    except Exception as e:
        logging.error(f"Error processing bacterium {i}: {e}")
        return None


def get_bacteria_list(img: np.array, mask_original: np.array, options: dict, pool=None, close_pool=True) -> list:
    if pool is None:
        logging.info("Making a pool for multithreading - this only happens once!")
        pool = ThreadPoolExecutor()  # Create executor outside a with block
        logging.info("Pool created!")
    else:
        logging.info("Using existing pool for multithreading")
    mask = label(mask_original)
    
    all_bacteria = []

    if len(img.shape) == 2:
        mask = clear_border(mask)  # Clear border to avoid issues with edge bacteria
        n_cells = np.max(mask)
        unique_values = np.unique(mask)
        # remove 0
        unique_values = unique_values[unique_values > 0]
        # args = [(img, mask, j, options) for j in range(1, n_cells + 1)]
        args = [(img, mask, j, options) for j in unique_values]

        # # single threaded version for debugging
        # if len(args) < 50:
        #     logging.info("Dataset is small. Using single-threaded processing.")
        #     with tqdm(total=len(args)) as pbar:
        #         for arg in args:
        #             result = _process_bacterium(arg)
        #             all_bacteria.append(result)
        #             pbar.update(1)
        # else:
            
        start_time = time.perf_counter()
        # Submit tasks and collect results
        futures = [pool.submit(_process_bacterium, arg) for arg in args]
        with tqdm(total=len(args)) as pbar:
            for future in as_completed(futures):
                result = future.result()
                all_bacteria.append(result)
                pbar.update(1)
        end_time = time.perf_counter()
        logging.info(f"Time taken for multithreading: {end_time - start_time:.2f} seconds")

        # Remove None values
        all_bacteria = [x for x in all_bacteria if x is not None]
    else:
        for i in range(img.shape[0]):
            current_img = np.copy(img[i])
            current_mask = clear_border(np.copy(mask[i]))
            current_bacteria = get_bacteria_list(current_img, current_mask, options, pool=pool, close_pool=False)
            for bact in current_bacteria:
                bact.slice = i
            all_bacteria.extend(current_bacteria)

    if close_pool:
        if pool:
            pool.shutdown(wait=True)
    return all_bacteria


from lmfit import Model
import numpy as np


def super_gaussian(x: np.array, center: float, width: float, amplitude: float, order: float or int, offset: float or
    int) -> np.array:
    """
    Super Gaussian model (top hat) to be used for fitting phase contrast profiles.

    Parameters
    ----------
    x: np.array
        values at which the model is evaluated
    center: float
        center of the super gaussian
    width: float
        width of the super gaussian
    amplitude: float
        amplitude of the super gaussian
    order: int
        order of the super gaussian (makes it more or less top-hat shaped)
    offset: float
        offset of the super gaussian (background)

    Returns
    -------
    value: np.array
        value(s) of the super gaussian at x
    """

    value = np.exp(-2*((x - center) / width) ** (order)) * amplitude + offset
    if np.isnan(value).any():
        print(f'nan value found for x={x}, center={center}, width={width}, order={1}, amplitude={amplitude}')
    return value

def fit_top_hat_profile(x: np.array, y: np.array):
    """
    Fit the phase contrast profile using a super gaussian (top-hat) model.

    Parameters
    ----------
    x: np.array
        x (independent) values of the profile

    y: np.array
        y values of the profile - this is what gets fitted

    Returns
    -------
    result : lmfit object
        the result of the fitting process, see lmfit documentation for more details
    """
    # Invert the profile
    # y = invert_profile(y)

    # Get initial guess for the parameters
    initial_guess = get_initial_guess(x, y)

    # Create the model
    gmodel = Model(super_gaussian)

    # Set the parameters
    params = gmodel.make_params(center=initial_guess['center'],
                                width=initial_guess['width'],
                                amplitude=initial_guess['amplitude'],
                                order=4,
                                offset=initial_guess['offset'])

    # fix order
    params['order'].vary = False

    # Fit the model
    result = gmodel.fit(y, params, x=x, nan_policy='raise')
    return result


def get_initial_guess(x: np.array, y: np.array) -> dict:
    """
    Get the initial guess for super gaussian model.

    Parameters
    ----------
    x: np.array
        values at which the model is evaluated
    y: np.array
        measured phase contrast profile

    Returns
    -------
    initial_guess: dict
        initial guess for the parameters, in a dictionary
    """
    # Determine initial guess for the parameters
    amplitude_guess = np.max(y)
    offset_guess = np.min(y)

    # find location minimum in first half of the data
    min_loc = x[np.argmin(y[:len(y) // 2])]
    # find location of minimum in second half of the data
    max_loc = x[(np.argmin(y[len(y) // 2:]) + len(y) // 2)]

    center_guess = (max_loc + min_loc) / 2

    width_guess = (max_loc - min_loc) / 2

    # Store all the initial guesses in a dictionary
    initial_guess = {'center': center_guess, 'width': width_guess, 'amplitude': amplitude_guess, 'offset': offset_guess}

    return initial_guess

def invert_profile(profile: np.array) -> np.array:
    """
    Convenience function to invert the profile to make it more suitable for fitting.

    Parameters
    ----------
    profile: np.array
        the profile to be inverted

    Returns
    -------
    profile: np.array
        the inverted profile
    """
    # Invert the profile
    profile = np.max(profile) - profile

    return profile

def fit_phase_contrast_profile(x: np.array, y: np.array):
    """
    Fit the phase contrast profile using a super gaussian (top-hat) model.

    Parameters
    ----------
    x: np.array
        x (independent) values of the profile

    y: np.array
        y values of the profile - this is what gets fitted

    Returns
    -------
    result : lmfit object
        the result of the fitting process, see lmfit documentation for more details
    """
    # Invert the profile
    y = invert_profile(y)

    # Get initial guess for the parameters
    initial_guess = get_initial_guess(x, y)

    # Create the model
    gmodel = Model(super_gaussian)

    # Set the parameters
    params = gmodel.make_params(center=initial_guess['center'],
                                width=initial_guess['width'],
                                amplitude=initial_guess['amplitude'],
                                order=4,
                                offset=initial_guess['offset'])

    # fix order
    params['order'].vary = False
    params['width'].min = 0


    # Fit the model
    result = gmodel.fit(y, params, x=x, nan_policy='raise')

    return result



from os.path import join
from skimage.io import imsave

from micromorph.segmentation import run_omnipose
from micromorph.batch.batch_utilities import get_tif_list
from micromorph.bacteria import get_bacteria_list
from micromorph.measure360.measure360 import run_measure360, filter_measure360
from micromorph.batch.image_loading_utils import load_image, apply_optosplit
from micromorph.batch.export_functions import export_full_analysis


# Functions for scripting batch processing of data in a folder

def run_omnipose_on_folder(folder_name: str, model_name: str, gpu_option: bool = False, chans=None, verbose=False,
                           mask_filter_options=None, loading_method: str = 'PIL', split_axis=None, split_channel=1):
    """
    Run omnipose on all tif files in a folder.
    :param folder_name: folder containing the tif files
    :param model_name: name of the omnipose model to use
    :param gpu_option: whether omnipose should use the GPU or not
    :param chans: 
    :param verbose: whether to print progress or not
    :param mask_filter_options: options for filtering the masks
    :param loading_method: method to use for loading the images, default is PIL
    :param split_axis: axis to split the image on - if using optosplit
    :param split_channel: channel to split the image on, if using optosplit
    """
    # Get the name of all the tif files in the folder
    file_list = get_tif_list(folder_name)

    # Remove masks from the list
    file_list = [x for x in file_list if 'mask' not in x]

    if verbose:
        print('Found', len(file_list), 'files in folder ', folder_name)

    # Run omnipose on each file
    for file in file_list:
        img = load_image(join(folder_name, file), method=loading_method, verbose=False)

        masks, flows = run_omnipose(img, model_name, gpu_option, filter_options=mask_filter_options)

        save_path = folder_name + '/' + file.replace('.tif', '_mask.tif')

        if split_axis is not None:
            masks = apply_optosplit(masks, optosplit_axis=split_axis, optosplit_channel=split_channel)
            img = apply_optosplit(img, optosplit_axis=split_axis, optosplit_channel=split_channel)

        if len(img.shape) >= 3:
            imsave(save_path, masks, shape=(masks.shape[0],masks.shape[1], masks.shape[2]))
        else:
            imsave(save_path, masks)

        if verbose:
            print('Processed ', file)

    return


def run_analysis_on_folder(folder_name, options=dict()):
    """
    Run the analysis on all tif files in a folder.
    :param folder_name: folder containing the tif files
    :param options: dictionary of options for the analysis
    :return: list of bacteria data
    """
    # Load from options, or set defaults.
    save_flag = options.get('save', True)
    output_type = options.get('output_type', 'all')

    save_directory = folder_name

    loading_method = options.get('loading_method', 'PIL')
    split_axis = options.get('split_axis', None)
    split_channel = options.get('split_channel', 1)

    file_list = get_tif_list(folder_name)

    # Remove masks from the list
    file_list = [x for x in file_list if 'mask' not in x]

    bacteria_data = None

    data_full_folder = []

    for file in file_list:
        img = load_image(join(folder_name, file), method=loading_method, verbose=False,
                         optosplit_channel=split_channel, optosplit_axis=split_axis)

        mask = load_image(join(folder_name, file.replace('.tif', '_mask.tif')), method=loading_method, verbose=False,
                         optosplit_channel=split_channel, optosplit_axis=split_axis)

        output_name = join(save_directory, file.replace('.tif', '_analysis'))

        # Run the analysis
        bacteria_data = get_bacteria_list(img, mask, options=options)

        data_full_folder.extend(bacteria_data)

        if save_flag:
            # Save the analysis for each image
            export_full_analysis(bacteria_data, output_name, output_type, data_type='Bacteria')

    if save_flag:
        # Save the full analysis for the folder
        export_full_analysis(data_full_folder, join(save_directory, 'full_folder_analysis'), output_type, data_type='Bacteria')

    return bacteria_data


def run_measure360_on_folder(folder_name,
                             analysis_options=dict(),
                             export_options=dict(),
                             image_loading_options=dict(),
                             filter_options=dict()):

    # combine the other dictionaries
    options = analysis_options.copy()
    options.update(export_options)
    options.update(image_loading_options)
    options.update(filter_options)

    save_directory = analysis_options.get('save_directory', folder_name)

    save_flag = analysis_options.get('save', True)

    apply_filter = filter_options.get('apply_filter', True)
    filter_type = filter_options.get('filter_type', 'derivative')
    filter_settings = filter_options.get('filter_settings', [250])

    loading_method = image_loading_options.get('loading_method', 'PIL')
    split_axis = image_loading_options.get('split_axis', None)
    split_channel = image_loading_options.get('split_channel', 1)

    output_type = export_options.get('output_type', 'all')

    file_list = get_tif_list(folder_name)
    # Remove masks from the list
    file_list = [x for x in file_list if 'mask' not in x]

    data_full_folder = []

    for file in file_list:
        img = load_image(join(folder_name, file), method=loading_method, verbose=False,
                         optosplit_channel=split_channel, optosplit_axis=split_axis)

        mask = load_image(join(folder_name, file.replace('.tif', '_mask.tif')), method=loading_method, verbose=False,
                          optosplit_channel=split_channel, optosplit_axis=split_axis)

        output_name = join(save_directory, file.replace('.tif', '_measure360'))

        print("Analysing image: ", file)
        bacteria_data = run_measure360(img, mask, options)

        if apply_filter:
            for bacteria in bacteria_data:
                bacteria = filter_measure360(bacteria, filter_type, filter_settings)

        data_full_folder.append(bacteria_data)

        if save_flag:
            export_full_analysis(bacteria_data, output_name, output_type, data_type='Bacteria360')

    if save_flag:
        # Save the full analysis for the folder
        export_full_analysis(data_full_folder, join(save_directory, 'full_folder_measure360'), output_type,
                             data_type='Bacteria360')

    return



import json
from pathlib import Path


def get_default() -> dict:
    """
    :return: default options as a dictionary
    """
    filepath = Path(__file__).parent / "default_options.json"
    with open(filepath) as f:
        return json.load(f)


def analysis_options() -> dict:
    """
    Get the default analysis options.
    """
    default_options = get_default()
    analysis_options_dict = default_options['analysis_options']
    return analysis_options_dict


def export_options() -> dict:
    """
    Get the default export options.
    """
    default_options = get_default()
    export_options_dict = default_options['export_options']
    return export_options_dict


def segmentation_options() -> dict:
    """
    Get the default segmentation options.
    """
    default_options = get_default()
    segmentation_options_dict = default_options['segmentation_options']

    return segmentation_options_dict


def image_loading_options() -> dict:
    """
    Get the default image loading options.
    """
    default_options = get_default()
    image_loading_options_dict = default_options['image_loading_options']
    return image_loading_options_dict


def filter_options() -> dict:
    """
    Get the default filter options.
    """
    default_options = get_default()
    filter_options_dict = default_options['filter_options']
    return filter_options_dict


if __name__ == "__main__":
    path = Path(__file__).parent / "default_options.json"
    print(path)
    # print(segmentation_options())



from .batch_functions import (run_analysis_on_folder, run_omnipose_on_folder, run_measure360_on_folder)

__all__ = [
    "run_analysis_on_folder",
    "run_omnipose_on_folder",
    "run_measure360_on_folder"
]


import pickle
import numpy as np

def export_analysis_setting(setting, output_path, method='pickle'):
    """
    NOT YET ACTUALLY IMPLEMENTED IN WORKFLOW - work in progress.
    Export the analysis setting dictionary to a file, in either pickle or txt format.
    :param setting: Dictionary containing the analysis settings
    :param output_path: Path to save the file
    :param method: 'pickle', 'txt'or 'both'
    """
    if method == 'pickle':
        with open(output_path+'.pickle', 'wb') as handle:
            pickle.dump(setting, handle, protocol=pickle.HIGHEST_PROTOCOL)
    elif method == 'txt':
        with open(output_path+'.txt', 'w') as file:
            for key, value in setting.items():
                file.write('%s:%s\n' % (key, value))
    elif method == 'both':
        export_analysis_setting(setting, output_path, 'pickle')
        export_analysis_setting(setting, output_path, 'txt')
    return


def export_full_analysis(bacteria_data: list, output_path, method='all', data_type=None):
    # check if data_type is either Bacteria or Bacteria360

    if data_type is None:
        # throw an error and stop
        raise ValueError('data_type must be specified')
    elif data_type != 'Bacteria' and data_type != 'Bacteria360':
        # throw an error and stop
        raise ValueError('data_type must be either Bacteria or Bacteria360')


    if method == 'pickle':
        # pickle method returns the actual list
        with open(output_path+'.pickle', 'wb') as handle:
            pickle.dump(bacteria_data, handle, protocol=pickle.HIGHEST_PROTOCOL)
    elif method == 'csv':
        try:
            import pandas as pd
        except ImportError:
            print("Pandas is not installed. Please install pandas to export CSV files.")
            return

        if data_type == 'Bacteria':
            data_dictionary = get_results_dictionary(bacteria_data)
        elif data_type == 'Bacteria360':
            data_dictionary = get_360_results_dictionary(bacteria_data)

        # convert data_dictionary to a dataframe
        df = pd.DataFrame(data_dictionary)
        df.to_csv(output_path+'.csv', index=False)
    elif method == 'hdf5':
        if data_type == 'Bacteria':
            data_dictionary = get_results_dictionary(bacteria_data)
        elif data_type == 'Bacteria360':
            data_dictionary = get_360_results_dictionary(bacteria_data)

        export_bacteria_to_hdf5(output_path+'.hdf5', data_dictionary)
    elif method == 'all':
        export_full_analysis(bacteria_data, output_path, method='hdf5', data_type=data_type)
        export_full_analysis(bacteria_data, output_path, method='pickle', data_type=data_type)
        export_full_analysis(bacteria_data, output_path, method='csv', data_type=data_type)
    return


def get_results_dictionary(bacteria_data: list) -> dict:
    """
    Convenience function which takes a list of Bacteria objects and returns a dictionary with the data,
    which can then be used to save the results with further functions.

    Parameters
    ----------
    bacteria_data: list
        list of Bacteria objects

    Returns
    -------
    data_dictionary: dict
        dictionary with the data from the Bacteria objects

    Keys
    ----
    frame: list
        list of frame numbers, zero if there was only one frame, or no frame information
    centroid: list
        list of centroid coordinates, in pixels
    medial_axis: list
        list of medial axis coordinates, in pixels
    boundary: list
        list of boundary coordinates, in pixels
    width: list
        list of widths of the bacteria, in units
    all_widths: list
        list of all widths found for each bacterium, in units
    length: list
        list of lengths of the bacteria, in units
    area: list
        list of areas of the bacteria, in pixels
    bounding_box: list
        list of bounding box coordinates, in pixels
    orientation: list
        list of orientation of the bacteria, in degrees
    major_axis_length: list
        list of major axis lengths of the bacteria, in units
    minor_axis_length: list
        list of minor axis lengths of the bacteria, in units
    """

    centroid = []
    medial_axis = []
    boundary = []
    width = []
    all_widths = []
    length = []
    area = []
    bounding_box = []
    orientation = []
    major_axis_length = []
    minor_axis_length = []
    medial_axis_extended = []
    frame = []

    for bacteria in bacteria_data:
        try:
            centroid.append(np.array([bacteria.xc, bacteria.yc]))
        except AttributeError:
            centroid.append(np.array([np.nan, np.nan]))
        medial_axis.append(bacteria.medial_axis)
        medial_axis_extended.append(bacteria.medial_axis_extended)
        width.append(bacteria.width)
        all_widths.append(bacteria.all_widths)
        length.append(bacteria.length)
        area.append(bacteria.area)
        bounding_box.append(bacteria.bbox)
        orientation.append(bacteria.orientation)
        major_axis_length.append(bacteria.axis_major_length)
        minor_axis_length.append(bacteria.axis_minor_length)
        frame.append(bacteria.slice)

        try:
            boundary.append(bacteria.boundary)
        except AttributeError:
            boundary.append(np.array([np.nan, np.nan]))

    data_dictionary = {'frame': frame,
                       'centroid': centroid,
                       'medial_axis': medial_axis,
                       'boundary': boundary,
                       'width': width,
                       'all_widths': all_widths,
                       'length': length,
                       'area': area,
                       'bounding_box': bounding_box,
                       'orientation': orientation,
                       'major_axis_length': major_axis_length,
                       'minor_axis_length': minor_axis_length,
                       'medial_axis_extended': medial_axis_extended}

    return data_dictionary


def get_360_results_dictionary(bacteria_data: list) -> dict:
    """
    Convenience function which takes a list of Bacteria360 objects and returns a dictionary with the data.
    This can then be used to save the results with further functions.

    """

    centroid = []
    width_data = []
    width = []
    length = []
    area = []
    bounding_box = []
    frame = []
    n_lines = []
    magnitude = []

    for bacteria in bacteria_data:
        try:
            centroid.append(np.array([bacteria.xc, bacteria.yc]))
        except AttributeError:
            centroid.append(np.array([np.nan, np.nan]))
        try:
            width_data.append(bacteria.width_data)
        except AttributeError:
            width_data.append(np.nan)
        try:
            width.append(bacteria.width)
        except AttributeError:
            width.append(np.nan)
        try:
            length.append(bacteria.length)
        except AttributeError:
            length.append(np.nan)


        try:
            area.append(bacteria.area)
        except AttributeError:
            area.append(np.nan)
        try:
            bounding_box.append(bacteria.bbox)
        except AttributeError:
            bounding_box.append(np.array([np.nan, np.nan]))
        try:
            frame.append(bacteria.slice)
        except AttributeError:
            frame.append(np.nan)
        try:
            n_lines.append(bacteria.n_lines)
        except AttributeError:
            n_lines.append(np.nan)
        try:
            magnitude.append(bacteria.magnitude)
        except AttributeError:
            magnitude.append(np.nan)

    data_dictionary = {'frame': frame,
                       'centroid': centroid,
                       'width_data': width_data,
                       'width': width,
                       'length': length,
                       'area': area,
                       'bounding_box': bounding_box,
                       'n_lines': n_lines,
                       'magnitude': magnitude}

    return data_dictionary


def export_bacteria_to_hdf5(hdf5_path, data_dictionary):
    try:
        import h5py

        with h5py.File(hdf5_path, 'w') as f:
            for key in data_dictionary.keys():
                f.create_group(key)

                for i, data in enumerate(data_dictionary[key]):
                    if data is None:
                        data = np.nan
                    f[key].create_dataset(str(i), data=data)
    except ImportError:
        print("Something went wrong. Please make sure h5py is installed correctly.")
        return
        # raise ValueError('Error saving data to hdf5 file')



from qtpy.QtGui import QIntValidator
from qtpy.QtWidgets import QWidget, QLabel, QLineEdit, QFormLayout, QHBoxLayout, QComboBox

class LabelledIntField(QWidget):
    def __init__(self, title, initial_value=None):
        QWidget.__init__(self)

        self.label = QLabel()
        self.label.setText(title)
        self.label.setContentsMargins(0, 0, 0, 0)
        # self.label.setFixedWidth(100)
        # layout.addWidget(self.label)

        self.lineEdit = QLineEdit(self)
        self.lineEdit.setContentsMargins(0, 0, 0, 0)
        # self.lineEdit.setFixedWidth(40)
        self.lineEdit.setValidator(QIntValidator())
        if initial_value != None:
            self.lineEdit.setText(str(initial_value))

        layout = QFormLayout()
        self.setLayout(layout)
        layout.addRow(QLabel(title), self.lineEdit)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)

    def setLabelWidth(self, width):
        self.label.setFixedWidth(width)

    def setInputWidth(self, width):
        self.lineEdit.setFixedWidth(width)

    def getValue(self):
        return int(self.lineEdit.text())

class LabelledFloatField(QWidget):
    def __init__(self, title, initial_value=None):
        QWidget.__init__(self)

        self.label = QLabel()
        self.label.setText(title)
        self.label.setContentsMargins(0, 0, 0, 0)
        # self.label.setFixedWidth(100)
        # layout.addWidget(self.label)

        self.lineEdit = QLineEdit(self)
        self.lineEdit.setContentsMargins(0, 0, 0, 0)
        # self.lineEdit.setFixedWidth(40)
        self.lineEdit.setValidator(QIntValidator())
        if initial_value != None:
            self.lineEdit.setText(str(initial_value))

        layout = QFormLayout()
        self.setLayout(layout)
        layout.addRow(QLabel(title), self.lineEdit)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)

    def setLabelWidth(self, width):
        self.label.setFixedWidth(width)

    def setInputWidth(self, width):
        self.lineEdit.setFixedWidth(width)

    def getValue(self):
        return float(self.lineEdit.text())

class DropdownMenu(QWidget):
    def __init__(self, label, options):
        QWidget.__init__(self)
        layout = QHBoxLayout()

        self.label = QLabel()
        self.label.setText(label)

        self.cb = QComboBox()
        self.cb.addItems(options)

        layout.addWidget(self.label)
        layout.addWidget(self.cb)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)


    def updateMenu(self, options):
        self.cb.clear()
        self.cb.addItems(options)


    def returnSelected(self):
        return self.cb.currentText()


class _layerSelectionWidget(QWidget):
    def __init__(self, napari_viewer, label="Select layer", layer_type=None):
        super().__init__()
        self.viewer = napari_viewer
        self._type = layer_type

        self.getLayerList()
        # self.layer_list = self.viewer.layers
        #
        # if layer_type is not None:
        #     self.layer_list = [layer for layer in self.layer_list if isinstance(layer, self._type)]

        self.layer_selection = DropdownMenu(label, [layer.name for layer in self.layer_list])

        layout = QHBoxLayout()

        layout.addWidget(self.layer_selection)

        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.viewer.layers.events.connect(self.updateLayerList)

    def getLayerList(self):
        all_layers_list = self.viewer.layers
        self.layer_list = [layer for layer in all_layers_list if isinstance(layer, self._type)]

    def returnSelectedLayer(self):
        return self.layer_selection.returnSelected()

    def updateLayerList(self):
        self.getLayerList()
        self.layer_selection.updateMenu([layer.name for layer in self.layer_list])



def get_tif_list(folder_path):
    """
    Get a list of tiff files in a folder
    :param folder_path: string, path to the folder
    :return: list of strings, list of tiff files in the folder
    """
    import os
    tiff_list = [f for f in os.listdir(folder_path) if f.endswith('.tif')]
    return tiff_list


import numpy as np
from shapely.geometry import Point, Polygon
from scipy.interpolate import splrep, BSpline
from skimage.measure import profile_line
from skimage.morphology import skeletonize, medial_axis
from .utilities import (get_width_profile_lines,
                        get_boundary_coords, fix_coordinates_order, trace_axis, prune_short_branches)
import logging
from .phase_contrast_fitting import fit_phase_contrast_profile, fit_top_hat_profile
from .fluorescence_fitting import fit_ring_profile
from scipy.ndimage import uniform_filter1d

"""
Collection of functions to analyse the shape of bacteria. These are all wrappers that run multiple other functions, 
but generate the output required for the Bacteria class.
"""


def get_bacteria_length(medial_axis_extended: np.array, pxsize: float or int = 1.0) -> float:
    """
    Calculate the length of a segmented bacterium.

    Parameters
    ----------
    medial_axis_extended : np.array
        The extended medial axis of the bacterium, whose arclength is calculated.
    pxsize : float, optional
        The pixel size of the image. The final distance will be multiplied by this value (default is 1.0).

    Returns
    -------
    total_distance : float
        Total distance covered by the extended medial axis (in px).
    """

    # Get distance vectors
    distance_vectors = np.diff(medial_axis_extended, axis=0)
    distance_vectors = np.concatenate((np.array([[0, 0]]), distance_vectors))

    # From the vectors, calculate the sums
    distances_individual = np.sqrt(np.sum(distance_vectors ** 2, axis=1))

    # now get total distance
    total_distance = np.sum(distances_individual)*pxsize

    return total_distance

def get_bacteria_widths(img: np.array, med_ax: np.array, n_lines: int = 5, pxsize: float or int = 1.0,
                        psfFWHM: float or int = 0.250, fit_type: str or None = None,
                        line_magnitude: float or int = 20) -> np.array:
    """
    Calculate the width of a segmented bacterium.

    Parameters
    ----------
    img : np.array
        The original image.
    med_ax : np.array
        The medial axis which will be used to find profile lines to fit.
    n_lines : int, optional
        The number of lines to fit (default is 5).
    pxsize : float or int, optional
        The pixel size of the image (default is 1.0).
    psfFWHM : float or int, optional
        The FWHM of the PSF (default is 5).
    fit_type : str or None, optional
        The type of fit to use (default is None).
    line_magnitude : float or int, optional
        The length of the line used to measure the profile (default is 20).

    Returns
    -------
    all_widths: np.array
        Array of widths.
    """
    xh, yh, xl, yl = get_width_profile_lines(med_ax, n_points=n_lines, line_magnitude=line_magnitude)

    profile_points = np.column_stack((xh, yh, xl, yl))

    all_widths = np.array([])

    for points in profile_points:
        # img or bound_transform
        current_profile = profile_line(img, (points[1], points[0]), (points[3], points[2]))

        x = np.arange(0, len(current_profile)) * pxsize

        try:
            if fit_type == 'fluorescence':
                result = fit_ring_profile(x, current_profile, psfFWHM)
                width = result.params['R'] * 2
            elif fit_type == 'phase':
                result = fit_phase_contrast_profile(x, current_profile)
                width = result.params['width'].value * 2 * (np.log(2)/2) ** (0.25)
            elif fit_type == 'tophat':
                result = fit_top_hat_profile(x, current_profile)
                width = result.params['width'].value * 2 * (np.log(2)/2) ** (0.25)
            else:
                raise ValueError('Invalid fit type - please choose "fluorescence" or "phase".')

            all_widths = np.append(all_widths, width)
        except:
            # logging.info("Error fitting profile, skipping this one.")
            pass

    return all_widths

def get_bacteria_boundary(mask: np.array, boundary_smoothing_factor: int or None = 7) -> np.array:
    """
    Get the (smoothed) boundary of a segmented bacterium.

    Parameters
    ----------
    mask : np.array
        The binary image (with a single bacterium only).
    boundary_smoothing_factor : int or None, optional
        The number of frequencies to use for FFT smoothing (default is 7).

    Returns
    -------
    boundary: np.array
        XY coordinates of the boundary (smoothed).
    """

    # Get initial coordinates for boundary coords
    boundary_xy = get_boundary_coords(mask)
    # boundary_xy = fix_coordinates_order(boundary_xy) # Performance of this may be questionable.

    if boundary_smoothing_factor is None:
        return boundary_xy
    else:
        x = boundary_xy[:, 0]
        y = boundary_xy[:, 1]

        x_fft = np.fft.rfft(x)
        x_fft[boundary_smoothing_factor:] = 0

        y_fft = np.fft.rfft(y)
        y_fft[boundary_smoothing_factor:] = 0

        x_smoothed = np.fft.irfft(x_fft)
        y_smoothed = np.fft.irfft(y_fft)

        boundary_xy_smoothed = np.column_stack((x_smoothed, y_smoothed))
        return np.fliplr(boundary_xy_smoothed)

def get_medial_axis(mask: np.array) -> np.array:
    """
    Get the medial axis of a segmented bacterium.

    Parameters
    ----------
    mask : np.array
        The binary image (with a single bacterium only).

    Returns
    -------
    medial_axis : np.array
        The medial axis of the bacterium.
    """
    bacteria_skeleton = skeletonize(mask)

    # Get the medial axis
    bacteria_skeleton = prune_short_branches(bacteria_skeleton)

    # Get the medial axis
    medial_axis = trace_axis(bacteria_skeleton)

    return medial_axis


# def smooth_medial_axis(medial_axis: np.array, boundary: np.array, error_threshold: float = 0.05, max_iter: int = 50, spline_val: int = 1, n_spline_points = 500) -> np.array:
#     """
#     Smooth the medial axis of a segmented bacterium, using the boundary as a reference.

#     Parameters
#     ----------
#     medial_axis : np.array
#         The medial axis of the bacterium.
#     boundary : np.array
#         The boundary of the bacterium.
#     error_threshold : float, optional
#         The threshold for the error in the smoothing (default is 0.05).
#     max_iter : int, optional
#         The maximum number of iterations for the smoothing (default is 50).

#     Returns
#     -------
#     smoothed_medial_axis: np.array
#         The smoothed medial axis.
#     """
#     print("NEW VERSION!! - 2")
#     # Sanity check that all points in the medial axis are INSIDE the boundary.
#     # This is due to the fact that we use the smoothed boundary for this calculation, but the medial axis is
#     # calculated from the mask, which can lead to points extending outside the boundary



#     medial_axis = np.array([uniform_filter1d(medial_axis[:,0], size=5), uniform_filter1d(medial_axis[:,1], size=5)]).T

#     iter_n = 0
#     previous_error = np.inf
#     while iter_n < max_iter:
#         # logging.info(f"Starting iteration {iter_n}")
#         middle_points = []
#         all_points = []
#         perpendicular_points = []

#         for i in range(medial_axis.shape[0] - 1):
#             # calculate the angle between the first and second point of the medial axis
#             angle = np.arctan2(medial_axis[i + 1, 1] - medial_axis[i, 1], medial_axis[i + 1, 0] - medial_axis[i, 0])

#             # now find the angle perpendicular to this angle
#             angle_perpendicular = angle + np.pi / 2

#             # calculate the middle point between the first and second point of the medial axis
#             middle_point = (medial_axis[i + 1] + medial_axis[i]) / 2


#             # get the closest point on the boundary to the perpendicular line
#             distances = np.abs((boundary[:, 1] - middle_point[1]) * np.cos(angle_perpendicular) - (
#                         boundary[:, 0] - middle_point[0]) * np.sin(angle_perpendicular))


#             # get the index of the closest point
#             # FIXME: this is ugly, but it works
#             idx = np.argmin(distances)

#             closest_point = boundary[idx]

#             # get the second closest
#             distances[idx] = np.inf
#             idx = np.argmin(distances)
#             closest_point2 = boundary[idx]

#             perpendicular_points.append(closest_point)
#             perpendicular_points.append(closest_point2)

#             # get the middle point between the two closest points
#             middle_point = (closest_point + closest_point2) / 2
#             middle_points.append(middle_point)

#             all_points.append(medial_axis[i])
#             all_points.append(middle_point)
#             all_points.append(medial_axis[i + 1])

#         # convert to numpy array
#         all_points = np.array(all_points)

#         # smooth the points
#         x = all_points[:, 0]
#         y = all_points[:, 1]

#         all_points = np.array([uniform_filter1d(x, size=5), uniform_filter1d(y, size=5)]).T

#         # all_points = np.array([x_smooth_s(calculation_values), y_smooth_s(calculation_values)]).T

#         # check if there are any large deviations in the smoothed points
#         # if there are, we need to stop the smoothing
#         delta_x = np.abs(np.diff(all_points[:, 0]))
#         if np.max(delta_x) > 100:
#             # logging.info(f"Large deviation in x, stopping smoothing here. Delta is {np.max(delta_x)}")
#             break

#         if iter_n >= 0:
#             current_error = np.abs(np.mean(all_points - medial_axis))
#             logging.info(f"Current error: {current_error}, iteration: {iter_n}/{max_iter}")
#             print(f"Current error: {current_error}, iteration: {iter_n}/{max_iter}")
#             if current_error > previous_error:
#                 pass
#                 logging.info(f"Error increased, stopping smoothing here. iteration: {iter_n}")
#                 print(f"Error increased, stopping smoothing here. iteration: {iter_n}")
#                 # break
#             elif current_error < error_threshold:
#                 medial_axis = all_points
#                 logging.info("Error below threshold, stopping smoothing here.")
#                 print("Error below threshold, stopping smoothing here.")
#                 break
#             else:
#                 medial_axis = all_points
#                 previous_error = current_error

#         iter_n += 1
#     return medial_axis


def smooth_medial_axis(medial_axis: np.array, boundary: np.array, error_threshold: float = 0.05, max_iter: int = 50, spline_val: int = 1, spline_spacing = 0.25) -> np.array:
    """
    Smooth the medial axis of a segmented bacterium, using the boundary as a reference.

    Parameters
    ----------
    medial_axis : np.array
        The medial axis of the bacterium.
    boundary : np.array
        The boundary of the bacterium.
    error_threshold : float, optional
        The threshold for the error in the smoothing (default is 0.05).
    max_iter : int, optional
        The maximum number of iterations for the smoothing (default is 50).

    Returns
    -------
    smoothed_medial_axis: np.array
        The smoothed medial axis.
    """
    # Sanity check that all points in the medial axis are INSIDE the boundary.
    # This is due to the fact that we use the smoothed boundary for this calculation, but the medial axis is
    # calculated from the mask, which can lead to points extending outside the boundary

    boundary_polygon = Polygon(boundary)
    medial_axis = np.array([point for point in medial_axis if Point(point).within(boundary_polygon)])

    # smooth medial axis before starting
    # smooth the points
    x = medial_axis[:, 0]
    y = medial_axis[:, 1]

    # need to calculate arclength of the medial axis, so we can use it for smoothing
    distances = np.sqrt(np.sum(np.diff(medial_axis, axis=0) ** 2, axis=1))
    distances = np.concatenate((np.array([0]), distances))  # add 0 for the first point
    total_arclength_distance = np.cumsum(distances)  # cumulative sum to get the arclength

    # n_spline_points is such that each point should be spaced px_spacing apart
    n_spline_points = int(np.ceil(total_arclength_distance[-1] / spline_spacing))

    # we are fitting x and y independently of each other
    try:
        indexes = np.arange(0, len(x), 1)

        tck_s = splrep(indexes, x, s=len(x), k=spline_val)
        x_smooth_s = BSpline(*tck_s)

        tck_sy = splrep(indexes, y, s=len(y), k=spline_val)
        y_smooth_s = BSpline(*tck_sy)
    except:
        return medial_axis

    calculation_values = np.linspace(0, len(x) - 1, n_spline_points)

    medial_axis = np.array([x_smooth_s(calculation_values), y_smooth_s(calculation_values)]).T

    iter_n = 0
    previous_error = np.inf
    while iter_n < max_iter:
        # logging.info(f"Starting iteration {iter_n}")
        middle_points = []
        all_points = []
        perpendicular_points = []

        for i in range(medial_axis.shape[0] - 1):
            # calculate the angle between the first and second point of the medial axis
            angle = np.arctan2(medial_axis[i + 1, 1] - medial_axis[i, 1], medial_axis[i + 1, 0] - medial_axis[i, 0])

            # now find the angle perpendicular to this angle
            angle_perpendicular = angle + np.pi / 2

            # calculate the middle point between the first and second point of the medial axis
            middle_point = (medial_axis[i + 1] + medial_axis[i]) / 2


            # get the closest point on the boundary to the perpendicular line
            distances = np.abs((boundary[:, 1] - middle_point[1]) * np.cos(angle_perpendicular) - (
                        boundary[:, 0] - middle_point[0]) * np.sin(angle_perpendicular))
            


            # get the index of the closest point
            # FIXME: this is ugly, but it works
            idx = np.argmin(distances)

            closest_point = boundary[idx]

            # get the second closest
            distances[idx] = np.inf
            idx = np.argmin(distances)
            closest_point2 = boundary[idx]

            perpendicular_points.append(closest_point)
            perpendicular_points.append(closest_point2)

            # get the middle point between the two closest points
            middle_point = (closest_point + closest_point2) / 2
            middle_points.append(middle_point)

            all_points.append(medial_axis[i])
            all_points.append(middle_point)
            all_points.append(medial_axis[i + 1])

        # convert to numpy array
        all_points = np.array(all_points)
        all_points = np.array([point for point in all_points if Point(point).within(boundary_polygon)])


        # smooth the points
        x = all_points[:, 0]
        y = all_points[:, 1]

        # we are fitting x and y independently of each other
        indexes = np.arange(0, len(x), 1)

        tck_s = splrep(indexes, x, s=len(x), k=spline_val)
        x_smooth_s = BSpline(*tck_s)

        tck_sy = splrep(indexes, y, s=len(y), k=spline_val)
        y_smooth_s = BSpline(*tck_sy)

        calculation_values = np.linspace(0, len(x) - 1, n_spline_points)

        all_points = np.array([x_smooth_s(calculation_values), y_smooth_s(calculation_values)]).T

        # check if there are any large deviations in the smoothed points
        # if there are, we need to stop the smoothing
        delta_x = np.abs(np.diff(all_points[:, 0]))
        if np.max(delta_x) > 100:
            # logging.info(f"Large deviation in x, stopping smoothing here. Delta is {np.max(delta_x)}")
            break

        if iter_n >= 0:
            current_error = np.abs(np.mean(all_points - medial_axis))
            # logging.info(f"Current error: {current_error}, iteration: {iter_n}/{max_iter}")
            if current_error > previous_error:
                pass
                # logging.info("Error increased, stopping smoothing here.")
                # break
            elif current_error < error_threshold:
                medial_axis = all_points
                # logging.info("Error below threshold, stopping smoothing here.")
                break
            else:
                medial_axis = all_points
                previous_error = current_error

        iter_n += 1
    return medial_axis

def extend_medial_axis_roughly(medax, bnd, max_iter=500, min_distance_to_boundary=1, step_size=1):
    """
    Extend the medial axis to the boundary, without any smoothing.

    Parameters
    ----------
    medax : np.array
        The medial axis.
    bnd : np.array
        The boundary.
    max_iter : int, optional
        The maximum number of iterations (default is 50).
    min_distance_to_boundary : int, optional
        The minimum distance to the boundary (default is 1).
    step_size : int, optional
        The step size to take (default is 1).

    Returns
    -------
    extended_medial_axis : np.array
        The extended medial axis.
    """
    medax_ext = np.copy(medax)

    # Generate a polygon to use for checking if the points are inside the boundary
    boundary_polygon = Polygon(bnd)
    # medial_axis = np.array([point for point in medial_axis if Point(point).within(boundary_polygon)])
    
    # We need to check that there's at least two points in the medial axis
    if medax.shape[0] < 2:
        return medax_ext

    # Direction second to last to last point
    direction = np.arctan2(medax[-1, 1] - medax[-2, 1], medax[-1, 0] - medax[-2, 0])

    dist_to_bound = np.inf
    N = 0
    while N < max_iter: #dist_to_bound > min_distance_to_boundary and 
        step_vector = step_size * np.array([np.cos(direction), np.sin(direction)])
        next_point = medax_ext[-1, :] + step_vector
        distances = np.sqrt((next_point[0] - bnd[:, 0]) ** 2 + (next_point[1] - bnd[:, 1]) ** 2)
        
        # check if the next point is outside the boundary
        if Point(next_point[0], next_point[1]).within(boundary_polygon):
            # logging.info("Next point is inside. Continuing")
            medax_ext = np.vstack((medax_ext, next_point))  
        else:
            # logging.info("Next point is outside. Stopping")
            medax_ext = np.vstack((medax_ext, next_point))  
            break

        
        N += 1

    # FIXME: UGLY - probably needs a second function which is called twice, not this ...
    # Repeat in the opposite direction

    # Direction second to first point
    direction = np.arctan2(medax[1, 1] - medax[0, 1], medax[1, 0] - medax[0, 0])

    dist_to_bound = np.inf
    N = 0
    while N < max_iter: # dist_to_bound > min_distance_to_boundary and 
        step_vector = -step_size * np.array([np.cos(direction), np.sin(direction)]) * 2
        next_point = medax_ext[0, :] + step_vector
    #     dist_to_bound = np.min(np.sqrt((next_point[0] - bnd[:, 0]) ** 2 + (next_point[1] - bnd[:, 1]) ** 2))
    #     # check if the next point is outside the boundaryp
    #     logging.info(f"Next point: {next_point}, distance to boundary: {dist_to_bound}")
        # check if the next point is outside the boundary#
        if Point(next_point[0], next_point[1]).within(boundary_polygon):
            medax_ext = np.vstack((next_point, medax_ext))
        else:
            logging.debug("Next point is outside. Stopping")
            break

        N += 1

    # remove first and last point
    medax_ext = np.delete(medax_ext, 0, axis=0)
    medax_ext = np.delete(medax_ext, -1, axis=0)

    return medax_ext

def extend_medial_axis(medial_axis: np.array, boundary: np.array, max_iter: int = 50,
                       error_threshold: float = 0.05, min_distance_to_boundary: int = 1, step_size: float or int = 1) -> np.array:
    """
    Extend the medial axis to the boundary, using the boundary as a reference, then smooth it to get the final result.

    Parameters
    ----------
    medial_axis : np.array
        The medial axis.
    boundary : np.array
        The boundary.
    max_iter : int, optional
        The maximum number of iterations for extending the medial axis(default is 50).
    min_distance_to_boundary : int, optional
        The minimum distance to the boundary (default is 1). This determines when the extension stops.
    step_size : float or int, optional
        The step size to take when extending the medial axis (default is 1 px).

    Returns
    -------
    extended_medial_axis: np.array
        The extended and smoothed medial axis.
    """

    medial_axis_extended_roughly = extend_medial_axis_roughly(medial_axis, boundary, step_size=0.05)


    # medial_axis_extended = smooth_medial_axis(medial_axis_extended_roughly, boundary, error_threshold=0.0005)

    return medial_axis_extended_roughly



import numpy as np
from tifffile import imread
import skimage.io as io

def apply_optosplit(img, optosplit_axis=0, optosplit_channel=1,verbose=False):
    img_split = np.zeros(img.shape, dtype=img.dtype)
    if verbose:
        print('Splitting image')

    # check if the image is a stack
    if len(img.shape) == 3:
        if optosplit_channel == 1:
            if optosplit_axis == 0:
                img_split[:, 0:img.shape[0] // 2, :] = img[:, 0:img.shape[0] // 2, :]
            elif optosplit_axis == 1:
                img_split[:, :, 0:img.shape[1] // 2] = img[:, :, 0:img.shape[1] // 2]
        else:
            if optosplit_axis == 0:
                img_split[:, img.shape[0] // 2:, :] = img[:, img.shape[0] // 2:, :]
            elif optosplit_axis == 1:
                img_split[:, :, img.shape[1] // 2:] = img[:, :, img.shape[1] // 2:]
    else:
        if optosplit_channel == 1:
            # Only return half of the columns
            if optosplit_axis == 0:
                img_split[:, 0:img.shape[0] // 2] = img[:, 0:img.shape[0] // 2]
            elif optosplit_axis == 1:
                img_split[0:img.shape[1] // 2, :] = img[0:img.shape[1] // 2, :]
        else:
            if optosplit_axis == 0:
                img_split[:, img.shape[0] // 2:] = img[:, img.shape[0] // 2:]
            elif optosplit_axis == 1:
                img_split[img.shape[1] // 2:, :] = img[img.shape[1] // 2:, :]
    if verbose:
        print('Image split successfully')

    return img_split


def load_image(file_path, method='tifffile', optosplit_channel=1, optosplit_axis=None, verbose=False):
    if method == 'PIL':
        if verbose:
            print('Reading image using PIL')
        try:
            img = io.imread(file_path, plugin='pil')
            if verbose:
                print('Image read successfully')
        except:
            if verbose:
                print('Failed to read image using PIL. Returning None.')
            img = None
    elif method == 'tifffile':
        try:
            if verbose:
                print('Reading image using tifffile')
            img = imread(file_path)
        except:
            print('Could not read file.')
            print('Attempting to use alternative method...')
            try:
                img = io.imread(file_path, plugin='pil')
                print('Alternative method seems to have worked.')
            except:
                print('Failed again. Returning None.')
                img = None
    else:
        print('Method not recognised. Returning None.')
        img = None

    # Additional options can be added here.
    if optosplit_axis is not None:
        img_split = apply_optosplit(img, optosplit_axis=optosplit_axis, optosplit_channel=optosplit_channel, verbose=verbose)

        return img_split
    else:
        return img




"""Bacteria analysis module for micromorph."""

# Import the main Bacteria class
from .Bacteria import (
    Bacteria,
    get_bacteria_for_index,
    get_bacteria_list,
)

# Import fluorescence fitting functions
from .fluorescence_fitting import gaussian

# Define what gets exported with "from micromorph.bacteria import *"
__all__ = [
    # Main class
    "Bacteria",
    # Bacteria.py functions
    "get_bacteria_for_index",
    "get_bacteria_list",
    # # Shape analysis
    # "get_bacteria_length",
    # "get_bacteria_widths",
    # "get_bacteria_boundary",
    # "get_medial_axis",
    # "smooth_medial_axis",
    # "extend_medial_axis",
    # "extend_medial_axis_roughly",
    # # Utilities
    # "find_endpoints",
    # "find_furthest_point",
    # "find_closest_point",
    # "find_branchpoints",
    # "prune_short_branches",
    # "get_coords",
    # "trace_axis",
    # "fix_coordinates_order",
    # "get_boundary_coords",
    # "find_closest_boundary_to_axis",
    # "calculate_distance_along_line",
    # "get_width_profile_lines",
    # "apply_mask_to_image",
    # # Fluorescence fitting
    # "gaussian",
]



from lmfit import Model
import numpy as np
from scipy.signal import convolve

def gaussian(params: tuple or list, x: np.array) -> np.array:
    """
    Gaussian function with 1 peak.

    Parameters
    ----------
    params: tuple or list
        parameters of the gaussian function [centre, sigma]
    x: np.array
        values at which the model is evaluated

    Returns
    -------
    values: np.array
        value(s) of the gaussian at x
    """

    a0 = params[0]
    a1 = params[1]

    # values = np.exp(-(x - a0) ** 2 / (a1 ** 2)) / (a1 * np.sqrt(2 * np.pi))
    values = np.exp(-(x - a0) ** 2 / (a1 ** 2))
    return values


def ring_profile(x: np.array, 
                 psfFWHM: int or float, 
                 R: int or float, 
                 x0: int or float, 
                 offset: int or float,
                 amp: int or float) -> np.array:
    """
    This function produces a line profile of a septum using an explicit 'tilted circle' model.
    It is based off code written in MATLAB by Séamus Holden, described in [Whitley et al. 2021](
    https://www.nature.com/articles/s41467-021-22526-0) and hosted in 

    Parameters
    ----------
    x: np.array
        values at which the model is evaluated
    x0: float
        the centre of the ring
    R: float
        the radius of the ring
    psfFWHM: float
        the FWHM of the microscope PSF, in the unit of x_values
    offset: float
        the background value
    amp: float
        the amplitude of the ring profile

    Returns
    -------
    image_profile: np.array
        value(s) of the ring profile at x
    """
    # n = 10
    # x_original = np.copy(x) # new
    # x = np.linspace(np.min(x), np.max(x), len(x) * n) # new
    # Calculate the sigma (standard deviation) of the PSF
    sigma = psfFWHM / 2.35

    x_extended = np.insert(x, 0, 0)

    profile_integral = 2 * R * np.real(np.emath.arcsin((x_extended - x0) / R))

    image_profile = np.diff(profile_integral)

    # # add zero to the beginning of the array
    # image_profile = np.insert(image_profile, 0, 0)

    # Convolution with gaussian function
    gauss_profile = gaussian([0, sigma], profile_integral)
    # gauss_profile = np.insert(gauss_profile, 0, 0)
    image_profile = convolve(image_profile, gauss_profile, mode='same')

    # image_profile = image_profile / np.max(image_profile) # temp
    image_profile = image_profile * amp + offset

    # calculate center of mass of image_profile
    com_profile = np.average(x, weights=image_profile)

    index_x0 = np.argmin(np.abs(x - x0))
    index_com = np.argmin(np.abs(x - com_profile))

    # shift the profile to the center of mass
    image_profile = np.roll(image_profile, - index_x0 + index_com)

    # image_profile_original = np.copy(image_profile) # new
    # image_profile = image_profile[::n] # new
    # image_profile_interp = np.interp(x_original, x, image_profile)
    return image_profile





def get_initial_guess(x: np.array, y: np.array):
    """
    Get initial guess for ring profile model.

    Parameters
    ----------
    x: np.array
        values at which the model is evaluated
    y: np.array
        measured ring profile

    Returns
    -------
    initial_guess: dict
        initial guess for the parameters, in a dictionary
    """
    # Determine initial guess for the parameters
    amplitude_guess = np.max(y)
    bg_guess = np.min(y)

    # x0 is the middle position
    x0_guess = x[len(x) // 2]

    # find location minimum in first half of the data
    min_loc = x[np.argmax(y[:len(y) // 2])]
    # find location of maximum in second half of the data
    max_loc = x[(np.argmax(y[len(y) // 2:]) + len(y) // 2)]

    # print(max_loc - min_loc)
    width_guess = (max_loc - min_loc) / 2  # this could possibly be determined from the FWHM of the peak

    initial_guess = {'x0': x0_guess, 'R': width_guess, 'offset': bg_guess, 'amp': amplitude_guess}

    return initial_guess

def prepare_data(profile: np.array)-> np.array:
    """
    Prepare the data for fitting, normalising the profile.

    Parameters
    ----------
    profile: np.array
        the profile to be prepared

    Returns
    -------
    profile: np.array
        the normalised profile
    """
    # Normalise the profile
    profile = profile / np.max(profile)


    return profile

def fit_ring_profile(x: np.array, y: np.array, psfFWHM: float):
    """
    Fit the ring profile using a tilted circle model.

    Parameters
    ----------
    x: np.array
        values at which the model is evaluated
    y: np.array
        measured ring profile
    psfFWHM: float
        the FWHM of the microscope PSF, in the unit of x_values

    Returns
    -------
    result: lmfit.model.ModelResult
        the result of the fitting
    """
    # Prepare the data
    y = prepare_data(y)

    # upscale the data - helps with ring fitting
    n = 10
    x_upscaled = np.linspace(np.min(x), np.max(x), len(x) * n)
    y_upscaled = np.interp(x_upscaled, x, y)

    # Get initial guess for the parameters
    initial_guess = get_initial_guess(x_upscaled, y_upscaled)

    # Create the model
    gmodel = Model(ring_profile)

    # Set the parameters
    params = gmodel.make_params(x0=initial_guess['x0'],
                                R=initial_guess['R'],
                                offset=initial_guess['offset'],
                                amp=initial_guess['amp'],
                                psfFWHM=psfFWHM)

    # Fix the psfFWHM
    params['psfFWHM'].vary = False

    # Fit the model
    result = gmodel.fit(y_upscaled, params, x=x_upscaled)
    return result



import numpy as np
from scipy.ndimage import binary_hit_or_miss
from skimage.morphology import label
from skimage.segmentation import find_boundaries
from skimage.measure import find_contours
from scipy.interpolate import splrep, splev

"""
Collection of utility functions used by the shape analysis functions, mainly to do with  binary image processing.
"""


def find_endpoints(img: np.array) -> np.array:
    """
        A function to find the endpoints of a binary image containing a line (i.e., a medial axis).

        Parameters
        ----------
        img : np.array
            The binary image.

        Returns
        -------
        endpoints : np.array
            A binary image with only the endpoints of the line.
    """
    endpoints = np.zeros(np.shape(img))

    endpoint_options = np.array([[1, 0, 0],
                                 [0, 1, 0],
                                 [0, 0, 0],
                                 [0, 1, 0],
                                 [0, 1, 0],
                                 [0, 0, 0],
                                 [0, 0, 1],
                                 [0, 1, 0],
                                 [0, 0, 0],
                                 [0, 0, 0],
                                 [0, 1, 1],
                                 [0, 0, 0],
                                 [0, 0, 0],
                                 [0, 1, 0],
                                 [0, 0, 1],
                                 [0, 0, 0],
                                 [0, 1, 0],
                                 [0, 1, 0],
                                 [0, 0, 0],
                                 [0, 1, 0],
                                 [1, 0, 0],
                                 [0, 0, 0],
                                 [1, 1, 0],
                                 [0, 0, 0],
                                 ])

    endpoint_structures = np.reshape(endpoint_options, (8, 3, 3))

    for box in endpoint_structures:
        current_attempt = binary_hit_or_miss(img, box)

        endpoints = endpoints + current_attempt
    return endpoints


def find_furthest_point(point: np.array, candidates: np.array) -> tuple[np.array, int]:
    """
    Find the furthest point from a given point, from a list of candidates.


    Parameters
    ----------
    point : np.array
        A np.array with 2 columns and 1 row (x, y)
    candidates : np.array
        An np.array with 2 columns and N rows (N being the number of candidates)

    Returns
    -------
    closest_point : np.array
        The furthest point
    max_index : int
        The index of the furthest point in the candidates array
    """
    distances_xy = candidates - point

    distances = np.sum(np.square(distances_xy), 1)

    max_index = np.where(distances == np.max(distances))

    max_index = max_index[0][0]  # Ensures we take the first one if there's two points with same distance

    closest_point = candidates[max_index, :]

    return closest_point, max_index


def find_closest_point(point: np.array, candidates: np.array) -> tuple[np.array, int]:
    """
    Find the closest point to a given point, from a list of candidates.

    Parameters
    ----------
    point : np.array
        A np.array with 2 columns and 1 row (x, y)
    candidates : np.array
        An np.array with 2 columns and N rows (N being the number of candidates)

    Returns
    -------
    closest_point : np.array
        The closest point
    min_index : int
        The index of the closest point in the candidates array
    """

    distances_xy = candidates - point

    distances = np.sum(np.square(distances_xy), 1)

    min_index = np.where(distances == np.min(distances))

    min_index = min_index[0][0]  # Ensures we take the first one if there's two points with same distance

    closest_point = candidates[min_index, :]

    return closest_point, min_index

def find_branchpoints(skeleton: np.array) -> np.array:
    """
    Function to find the branchpoints of a binary image containing a skeleton.

    Parameters
    ----------
    skeleton : np.array
        The skeletonised binary image

    Returns
    -------
    branch_points : np.array
        A binary image with only the branchpoints of the skeleton
    """
    selems = list()
    selems.append(np.array([[0, 1, 0], [1, 1, 1], [0, 0, 0]]))
    selems.append(np.array([[1, 0, 1], [0, 1, 0], [1, 0, 0]]))
    selems.append(np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0]]))
    selems.append(np.array([[0, 1, 0], [1, 1, 0], [0, 0, 1]]))
    selems.append(np.array([[0, 0, 1], [1, 1, 1], [0, 1, 0]]))
    selems = [np.rot90(selems[i], k=j) for i in range(5) for j in range(4)]

    selems.append(np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]]))
    selems.append(np.array([[1, 0, 1], [0, 1, 0], [1, 0, 1]]))

    branch_points = np.zeros_like(skeleton, dtype=bool)
    for selem in selems:
        branch_points |= binary_hit_or_miss(skeleton, selem)

    return branch_points


def prune_short_branches(skeleton: np.array) -> np.array:
    """
    Function to find the longest branch in a skeleton.

    Parameters
    ----------
    skeleton : np.array
        The binary image, skeletonised

    Returns
    -------
    pruned_skeleton : np.array
        A binary image with only the longest branch of the skeleton
    """
    if not skeleton.dtype == bool:
        skeleton = skeleton.astype(bool)

    branch_points = find_branchpoints(skeleton)
    separated_branches = np.copy(skeleton) ^ branch_points

    labelled_branches = label(separated_branches)

    unique_labels = np.unique(labelled_branches)

    # remove zero, as it's the background
    unique_labels = unique_labels[1:]

    # find the highest occurence
    max_occurence = 0
    max_label = 0
    for n in unique_labels:
        occurence = np.sum(labelled_branches == n)
        if occurence > max_occurence:
            max_occurence = occurence
            max_label = n

    pruned_skeleton = labelled_branches == max_label

    return pruned_skeleton


def get_coords(binary_image: np.array) -> np.array:
    """
    A function to get the coordinates of the non-zero pixels in a binary image.

    Parameters
    ----------
    binary_image : np.array
        A binary image

    Returns
    -------
    xy : np.array
        An np.array with 2 columns and N rows (N being the number of non-zero pixels)
    """

    xy = np.fliplr(np.asarray(np.where(binary_image)).T)
    return xy


def trace_axis(medial_axis_image: np.array) -> np.array:
    """
    Trace the medial axis, returning an ordered array of points from one end to the other. The first endpoint is defined as the most leftmost of the endpoints.

    Parameters
    ----------
    medial_axis_image : np.array
        The binary image with the medial axis

    Returns
    -------
    medial_axis : np.array
        An np.array with 2 columns and N rows (N being the number of points in the medial axis)
    """
    # Want to get medial axis in ordered point list.

    # Get coordinates (unordered) of the medial axis
    medial_axis_non_ordered = get_coords(medial_axis_image)

    # Grab endpoints of the medial axis line, and get their coordinates
    endpoints_image = find_endpoints(medial_axis_image)
    endpoints = get_coords(endpoints_image)

    # Pick one endpoint to start ordering from
    # This is arbitrary, so let's pick point that is the leftmost
    chosen_endpoint_index = np.where(endpoints[:, 0] == np.min(endpoints[:, 0]))
    chosen_endpoint_index = chosen_endpoint_index[0][0]

    # Create a new array where the ordered points will be stored
    medial_axis = np.zeros(medial_axis_non_ordered.shape)

    # Assign first point, and remove it from medial_axis_non_ordered
    medial_axis[0, :] = endpoints[chosen_endpoint_index, :]

    index_to_delete = np.where((medial_axis_non_ordered[:, 0] == endpoints[chosen_endpoint_index, 0]) &
                               (medial_axis_non_ordered[:, 1] == endpoints[chosen_endpoint_index, 1]))

    medial_axis_non_ordered = np.delete(medial_axis_non_ordered, index_to_delete, 0)

    for i, point in enumerate(medial_axis, start=1):
        previous_point = medial_axis[i - 1, :]

        if medial_axis_non_ordered.shape[0] > 1:
            closest_point, index = find_closest_point(previous_point, medial_axis_non_ordered)

            medial_axis[i, :] = closest_point

            # Delete point from pool
            medial_axis_non_ordered = np.delete(medial_axis_non_ordered, index, 0)
        else:
            medial_axis[i, 0] = medial_axis_non_ordered[0][0]
            medial_axis[i, 1] = medial_axis_non_ordered[0][1]
            break

    return medial_axis


def fix_coordinates_order(coordinates: np.array) -> np.array:
    """
    This function fixes the coordinate order in a list of points (usually extracted from a binary image).
    In shorts, it picks a point and then finds the closest point to it, and so on.

    Parameters
    ----------
    coordinates : np.array
        An np.array with 2 columns and N rows (N being the number of points)

    Returns
    -------
    ordered_coordinates : np.array
        An np.array with 2 columns and N rows (N being the number of points)
    """
    ordered_coordinates = np.zeros(coordinates.shape)

    # Set first coordinate of the boundary
    ordered_coordinates[0, :] = coordinates[0, :]
    coordinates = np.delete(coordinates, 0, 0)

    for i, point in enumerate(ordered_coordinates, start=1):
        previous_point = ordered_coordinates[i - 1, :]
        if coordinates.shape[0] > 1:
            closest_point, index = find_closest_point(previous_point, coordinates)

            ordered_coordinates[i, :] = closest_point

            coordinates = np.delete(coordinates, index, 0)
        else:
            candidate = np.array([[coordinates[0][0], coordinates[0][1]]])

            distance_to_candidate = np.sqrt(np.sum((candidate - previous_point) ** 2))

            if distance_to_candidate < 3:
                ordered_coordinates[i, 0] = coordinates[0][0]
                ordered_coordinates[i, 1] = coordinates[0][1]
            else:
                ordered_coordinates = np.delete(ordered_coordinates, i, 0)
            break

    return ordered_coordinates


def get_boundary_coords(binary_image: np.array) -> np.array:
    """
    Runs the find boundary function from skimage, and then extracts the coordinates of the boundary.

    Parameters
    ----------
    binary_image : np.array
        A binary image

    Returns
    -------
    boundary_xy : np.array
        An np.array with 2 columns and N rows (N being the number of boundary pixels). The points will not be ordered.
    """

    # Get boundaries from binary image
    # boundary_img = find_boundaries(binary_image, connectivity=4, mode='inner', background=0)
    boundary_xy = find_contours(binary_image)[0]

    # # ...and extract coordinates
    # boundary_xy = get_coords(boundary_img)
    return boundary_xy


def find_closest_boundary_to_axis(medial_axis_coords: np.array, boundary_coords: np.array,
                                  start_position: int) -> np.array:
    """
    A function to find the closest point on the boundary to the medial axis.

    Parameters
    ----------
    medial_axis_coords : np.array
        An np.array with 2 columns and N rows (N being the number of points in the medial axis).
    boundary_coords : np.array
        An np.array with 2 columns and M rows (M being the number of points in the boundary).
    start_position : int
        Either 1 or -1. If 1, the medial axis will be ordered from the first point to the last. If -1, the medial axis will be ordered from the last point to the first.

    Returns
    -------
    closest_on_boundary : np.array
        An np.array with 2 columns and 1 row, the coordinates of the closest point on the boundary.
    """
    # Start position can be either 1 or -1
    if start_position == 1:
        pass
    else:
        medial_axis_coords = np.flipud(medial_axis_coords)

    endpoint = medial_axis_coords[0, :]

    if len(medial_axis_coords) > 10:
        points_of_interest = medial_axis_coords[1:10, :]
    else:
        points_of_interest = medial_axis_coords[1::, :]

    search_vectors = points_of_interest - endpoint
    # This is to fix bug that happens when medial_axis_coords is only two points.
    if len(search_vectors) == 2:
        search_vectors = np.reshape(search_vectors, [1, 2])

    search_angles = np.arctan2(search_vectors[:, 1], search_vectors[:, 0])

    mean_search_angle = np.mean(search_angles)

    # Get angles from endpoint to all the boundary points.
    vectors_to_boundary_points = endpoint - boundary_coords
    angles_to_boundary_points = np.arctan2(vectors_to_boundary_points[:, 1], vectors_to_boundary_points[:, 0])

    selected_points = np.where((angles_to_boundary_points > mean_search_angle - 0.3) &
                               (angles_to_boundary_points < mean_search_angle + 0.3))

    # Filter the boundary points
    filtered_boundary_points = boundary_coords[selected_points, :]

    # closest_on_boundary = find_closest_point(endpoint, filtered_boundary_points)
    closest_on_boundary = find_furthest_point(endpoint, filtered_boundary_points)
    closest_on_boundary = np.array(closest_on_boundary[0][0])
    return closest_on_boundary


def calculate_distance_along_line(coordinates: np.array) -> np.array:
    """
    Function to calculate the distance along a line defined by a set of coordinates.

    Parameters
    ----------
    coordinates : np.array
        An np.array with 2 columns and N rows (N being the number of points in the line).

    Returns
    -------
    distances_cumulative : np.array
        An np.array with 1 column and N rows (N being the number of points in the line).
    """
    # Get distance vectors
    distance_vectors = np.diff(coordinates, axis=0)
    distance_vectors = np.concatenate((np.array([[0, 0]]), distance_vectors))

    # From the vectors, calculate the sums
    distances_individual = np.sqrt(np.sum(distance_vectors ** 2, axis=1))
    distances_cumulative = np.cumsum(distances_individual)

    return distances_cumulative


def get_width_profile_lines(medial_axis: np.array, n_points: int = 3, line_magnitude: int or float = 10):
    """
    Function to get the lines that will be used to calculate the width of the bacterium through fitting.

    Parameters
    ----------
    medial_axis : np.array
        An np.array with 2 columns and N rows (N being the number of points in the medial axis).
    n_points : int
        The number of points to sample along the medial axis.
    line_magnitude : int or float
        The length of the lines that will be perpendicular to the medial axis.

    Returns
    -------
    x_high, y_high, x_low, y_low : np.arrays
        np.arrays with 1 column and N rows (N being the number of points in the medial axis).
    """

    # Find slope of medial axis
    delta_vectors = np.diff(medial_axis, axis=0)
    parallel_angles = np.arctan2(delta_vectors[:, 1], delta_vectors[:, 0])
    perpendicular_angles = parallel_angles + np.deg2rad(90)

    # Remove first point from the medial axis, to make it same length as other vectors
    medial_axis = np.delete(medial_axis, 0, axis=0)

    # Get arclength curves
    distances = calculate_distance_along_line(medial_axis)

    # Get spline fit of distances vs x
    x_coords = medial_axis[:, 0]
    y_coords = medial_axis[:, 1]


    if len(distances) > 3:
        x_spline = splrep(distances, x_coords)
        y_spline = splrep(distances, y_coords)
        angles_spline = splrep(distances, perpendicular_angles)

        sampling_distances = np.linspace(0, distances[-1], n_points, endpoint=False)

        x_points = []
        y_points = []
        angles = []

        for val in sampling_distances:
            x_points.append(splev(val, x_spline))
            y_points.append(splev(val, y_spline))
            angles.append(splev(val, angles_spline))
    else:
        x_points = x_coords
        y_points = y_coords
        angles = perpendicular_angles

    x_ref = x_points
    y_ref = y_points
    perpendicular_angles = angles

    # Create lines 1 distance_line away from medial axis point, in each direction
    dx_perp = np.cos(perpendicular_angles) * line_magnitude / 2
    dy_perp = np.sin(perpendicular_angles) * line_magnitude / 2

    x_high = x_ref + dx_perp
    y_high = y_ref + dy_perp

    x_low = x_ref - dx_perp
    y_low = y_ref - dy_perp

    return x_high, y_high, x_low, y_low


def apply_mask_to_image(img: np.array, mask: np.array, method: str = 'min') -> np.array:
    """
    A function to apply a dilation to a mask, and then apply the mask to the image.
    This makes all the points outside the region of interest black.

    Parameters
    ----------
    img : np.array
        The image to be masked
    mask : np.array
        The mask to be applied to the image

    Returns
    -------
    img_masked : np.array
        The masked image
    """

    mask_inverted = np.invert(np.copy(mask))

    img_masked = np.copy(img)
        
    # Below is to avoid problems with different data types
    if mask_inverted.dtype == 'bool':
        idx = mask_inverted == True
    else:
        idx = mask_inverted == 255

    # img_masked[idx] = np.mean(img)
    if method == 'min':
        img_masked[idx] = np.min(img)
    elif method == 'max':
        img_masked[idx] = np.max(img)
    elif method == 'mean':
        img_masked[idx] = np.mean(img)
    else:
        # raise a warning
        Warning('Method not recognised, using min instead')
        img_masked[idx] = np.min(img)

    return img_masked



from qtpy.QtWidgets import (QPushButton, QWidget, QGridLayout,
                            QVBoxLayout, QCheckBox)
from ._gui_utils import LabelledIntField, _layerSelectionWidget
from micromorph.segmentation import filter_mask


class MaskFilterWidget(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()

        if napari_viewer:
            self.viewer = napari_viewer
            # Create dropdown menu for layer selection
            self.layer_selection_dropdown = _layerSelectionWidget(self.viewer, layer_type=napari.layers.labels.labels.Labels)

        # Checkboxes for filtering
        self.min_area_checkbox = QCheckBox()
        self.max_area_checkbox = QCheckBox()
        self.min_length_checkbox = QCheckBox()
        self.max_length_checkbox = QCheckBox()
        self.min_width_checkbox = QCheckBox()
        self.max_width_checkbox = QCheckBox()

        # Create fields for the user to input parameters
        self.min_area = LabelledIntField("Minimum area", 0)
        self.max_area = LabelledIntField("Maximum area", 10000)

        self.min_length = LabelledIntField("Minimum length", 0)
        self.max_length = LabelledIntField("Maximum length", 1000)

        self.min_width = LabelledIntField("Minimum width", 0)
        self.max_width = LabelledIntField("Maximum width", 1000)

        if napari_viewer:
            # Create buttons
            self.run_button = QPushButton("Apply filter")

        # Layout
        main_layout = QVBoxLayout()
        layout = QGridLayout()
        main_layout.addLayout(layout)

        if napari_viewer:
            layout.addWidget(self.layer_selection_dropdown, 0, 0, 1, 2)

        layout.addWidget(self.min_area, 1, 0, 1, 1)
        layout.addWidget(self.max_area, 2, 0, 1, 1)
        layout.addWidget(self.min_length, 3, 0, 1, 1)
        layout.addWidget(self.max_length, 4, 0, 1, 1)
        layout.addWidget(self.min_width, 5, 0, 1, 1)
        layout.addWidget(self.max_width, 6, 0, 1, 1)

        layout.addWidget(self.min_area_checkbox, 1, 1, 1, 1)
        layout.addWidget(self.max_area_checkbox, 2, 1, 1, 1)
        layout.addWidget(self.min_length_checkbox, 3, 1, 1, 1)
        layout.addWidget(self.max_length_checkbox, 4, 1, 1, 1)
        layout.addWidget(self.min_width_checkbox, 5, 1, 1, 1)
        layout.addWidget(self.max_width_checkbox, 6, 1, 1, 1)

        if napari_viewer:
            layout.addWidget(self.run_button, 7, 0, 1, 2)

        self.setLayout(main_layout)

        if napari_viewer:
            # Connect operations
            self.run_button.clicked.connect(self.run_filter_on_selected_mask)

    def run_filter_on_selected_mask(self):
        # Get selected layer
        layer_name = self.layer_selection_dropdown.layer_selection.returnSelected()

        label_data = self.viewer.layers[layer_name].data

        filter_options = self.get_filter_options()

        filtered_label_data = filter_mask(label_data, options=filter_options)

        # create a new labels layer
        results_name = "Filtered labels of " + layer_name
        self.viewer.add_labels(filtered_label_data, name=results_name)

    def get_filter_options(self):
        if self.min_area_checkbox.isChecked():
            min_area = self.min_area.getValue()
        else:
            min_area = None

        if self.max_area_checkbox.isChecked():
            max_area = self.max_area.getValue()
        else:
            max_area = None

        if self.min_length_checkbox.isChecked():
            min_length = self.min_length.getValue()
        else:
            min_length = None

        if self.max_length_checkbox.isChecked():
            max_length = self.max_length.getValue()
        else:
            max_length = None

        if self.min_width_checkbox.isChecked():
            min_width = self.min_width.getValue()
        else:
            min_width = None

        if self.max_width_checkbox.isChecked():
            max_width = self.max_width.getValue()
        else:
            max_width = None

        filter_options = {'min_area': min_area,
                          'max_area': max_area,
                          'min_length': min_length,
                          'max_length': max_length,
                          'min_width': min_width,
                          'max_width': max_width}

        return filter_options
    